"""Pre and post hooks on agents."""

from __future__ import annotations

__all__ = [
    "asset_merge_factory",
    "clean",
    "housekeeping_factory",
    "merge_assets",
    "new_assets_only",
    "noop",
    "old_assets_only",
    "register_final_asset_transform",
    "register_initial_asset_transform",
]

from collections.abc import Mapping, MutableMapping
from typing import Callable

from xarray import Dataset

from muse.agents import Agent
from muse.registration import registrator

INITIAL_ASSET_TRANSFORM: MutableMapping[str, Callable] = {}
""" Transform at the start of each step. """
FINAL_ASSET_TRANSFORM: MutableMapping[str, Callable] = {}
""" Transform at the end of each step, including new assets. """


def housekeeping_factory(settings: str | Mapping = "clean") -> Callable:
    """Returns a function for performing initial housekeeping.

    For instance, remove technologies with no capacity now or in the future.
    Available housekeeping functions should be registered with
    :py:func:`@register_initial_asset_transform<register_initial_asset_transform>`.
    """
    from muse.agents import AbstractAgent

    if isinstance(settings, str):
        name = settings
        params: Mapping = {}
    else:
        params = {k: v for k, v in settings.items() if k != "name"}
        name = settings["name"]

    transform = INITIAL_ASSET_TRANSFORM[name]

    def initial_assets_transform(agent: AbstractAgent, assets: Dataset) -> Dataset:
        return transform(agent, assets, **params)

    return initial_assets_transform


def asset_merge_factory(settings: str | Mapping = "new") -> Callable:
    """Returns a function for merging new investments into assets.

    Available merging functions should be registered with
    :py:func:`@register_final_asset_transform<register_final_asset_transform>`.
    """
    """Returns a function for performing initial housekeeping.

    For instance, remove technologies with no capacity now or in the future.
    Available housekeeping functions should be registered with
    :py:func:`@register_initial_asset_transform<register_initial_asset_transform>`.
    """
    if isinstance(settings, str):
        name = settings
        params: Mapping = {}
    else:
        params = {k: v for k, v in settings.items() if k != "name"}
        name = settings["name"]

    transform = FINAL_ASSET_TRANSFORM[name]

    def final_assets_transform(old_assets: Dataset, new_assets):
        return transform(old_assets, new_assets, **params)

    final_assets_transform.__name__ = name
    return final_assets_transform


@registrator(registry=INITIAL_ASSET_TRANSFORM, loglevel="info")
def register_initial_asset_transform(
    function: Callable[[Agent, Dataset], Dataset],
) -> Callable:
    """Decorator to register a function for cleaning or transforming assets.

    The transformation is applied at the start of each iteration. It any function which
    take an agent and assets as input and any number of keyword arguments, and returns
    the transformed assets. The agent should not be modified. It is only there to query
    the current year, the region, etc.
    """
    return function


@registrator(registry=FINAL_ASSET_TRANSFORM, loglevel="info")
def register_final_asset_transform(
    function: Callable[[Dataset, Dataset], Dataset],
) -> Callable:
    """Decorator to register a function to merge new investments into current assets.

    The transform is applied a the very end of the agent iteration. It can be any
    function which takes as input the current set of assets, the new assets, and any
    number of keyword arguments. The function must return a "merge" of the two assets.

    For instance, the new assets could completely replace the old assets
    (:py:func:`new_assets_only`), or they could be summed to the old assets
    (:py:func:`merge_assets`).
    """
    from functools import wraps

    @wraps(function)
    def decorated(old_assets: Dataset, new_assets: Dataset) -> Dataset:
        result = function(old_assets, new_assets)
        # missing values -> NaN -> integers become floats
        for variable in set(result.variables).intersection(old_assets.variables):
            result[variable] = result[variable].astype(old_assets[variable].dtype)
        result = result.drop_vars(set(result.coords) - set(old_assets.coords))
        return result

    return decorated


@register_initial_asset_transform(name="default")
def noop(agent: Agent, assets: Dataset) -> Dataset:
    """Return assets as they are."""
    return assets


@register_initial_asset_transform
def clean(agent: Agent, assets: Dataset) -> Dataset:
    """Removes empty assets."""
    from muse.utilities import clean_assets

    assets_clean = clean_assets(assets, agent.year)

    # Will run into issues if there are no assets, so just return the original array
    # TODO: need a more robust solution for this problem
    if len(assets_clean.asset) == 0:
        return assets
    return assets_clean


@register_final_asset_transform(name="new")
def new_assets_only(old_assets: Dataset, new_assets: Dataset) -> Dataset:
    """Returns newly invested assets and ignores old assets."""
    return new_assets


@register_final_asset_transform(name="old")
def old_assets_only(old_assets: Dataset, new_assets: Dataset) -> Dataset:
    """Returns old assets and ignores newly invested assets."""
    return old_assets


@register_final_asset_transform(name="merge")
def merge_assets(old_assets: Dataset, new_assets: Dataset) -> Dataset:
    """Adds new assets to old along asset dimension.

    New assets are assumed to be nonequivalent to any old_assets. Indeed,
    it is expected that the asset dimension does not have coordinates (i.e.
    it is a combination of coordinates, such as technology and installation
    year).

    After merging the new assets, quantities are back-filled along the year
    dimension. Further missing values (i.e. future years the old_assets
    did not take into account) are set to zero.
    """
    from logging import getLogger

    from muse.utilities import merge_assets

    assert "asset" not in old_assets

    if "asset" not in new_assets.dims and "replacement" in new_assets.dims:
        new_assets = new_assets.rename(replacement="asset")
    if len(new_assets.capacity) == 0:
        getLogger(__name__).critical("there are no new assets")
        return old_assets
    return merge_assets(old_assets, new_assets)



"""Sinks where output quantities can be stored.

Sinks take as argument a DataArray and store it somewhere. Additionally they
take a dictionary as argument. This dictionary will always contains the items
('quantity', 'sector', 'year') referring to the name of the quantity, the name
of the calling sector, the current year. They may contain additional parameters
which depend on the actual sink, such as 'filename'.

Optionally, a description of the storage (filename, etc) can be returned.

The signature of a sink is:

.. code-block:: python

    @register_output_sink(name="netcfd")
    def to_netcfd(quantity: DataArray, config: Mapping) -> Optional[Text]:
        pass
"""

from __future__ import annotations

from collections.abc import Mapping, MutableMapping, Sequence
from typing import (
    Any,
    Callable,
    Optional,
    Union,
    cast,
)

import pandas as pd
import xarray as xr
from mypy_extensions import KwArg

from muse.registration import registrator

OUTPUT_SINK_SIGNATURE = Callable[
    [Union[xr.DataArray, pd.DataFrame], int, KwArg(Any)], Optional[str]
]
"""Signature of functions used to save quantities."""

OUTPUT_SINKS: MutableMapping[str, OUTPUT_SINK_SIGNATURE | Callable] = {}
"""Stores a quantity somewhere."""


def factory(parameters: Mapping, sector_name: str = "default") -> Callable:
    from functools import partial
    from inspect import isclass
    from pathlib import Path

    from muse.outputs.sinks import OUTPUT_SINKS

    config = dict(**parameters)
    config.pop("quantity", None)

    def normalize(
        params: Mapping | None, filename: str | None = None
    ) -> MutableMapping:
        if isinstance(params, Mapping):
            params = dict(**params)
        elif isinstance(params, str):
            params = dict(name=params)
        else:
            suffix = config.get("suffix", Path(filename).suffix if filename else "csv")
            if not suffix:
                suffix = "csv"
            params = dict(name=suffix)
        if "aggregate" in params:
            params["name"] = "aggregate"
            params["final_sink"] = dict(
                sink=normalize(params.pop("aggregate"), filename)
            )
        return params

    params = normalize(config.pop("sink", None), config.get("filename", None))
    sink_name = params.pop("name")

    if len(set(params).intersection(config)) != 0:
        raise ValueError("duplicate settings in output section")
    params.update(config)
    params["sector"] = sector_name.lower()
    if sink_name[0] == ".":
        sink_name = sink_name[1:]
    sink = OUTPUT_SINKS[sink_name]
    if isclass(sink):
        return sink(**params)  # type: ignore
    else:
        return partial(sink, **params)


@registrator(registry=OUTPUT_SINKS, loglevel=None)
def register_output_sink(function: OUTPUT_SINK_SIGNATURE = None) -> Callable:
    """Registers a function to save quantities."""
    assert function is not None
    return function


def sink_to_file(suffix: str):
    """Simplifies sinks to files.

    The decorator takes care of figuring out the path to the file, as well as trims the
    configuration dictionary to include only parameters for the sink itself. The
    decorated function returns the path to the output file.
    """
    from functools import wraps
    from logging import getLogger
    from pathlib import Path

    from muse.defaults import DEFAULT_OUTPUT_DIRECTORY

    def decorator(function: Callable[[pd.DataFrame | xr.DataArray, str], None]):
        @wraps(function)
        def decorated(
            quantity: pd.DataFrame | xr.DataArray, year: int, **config
        ) -> Path:
            params = config.copy()
            filestring = str(
                params.pop(
                    "filename", "{default_output_dir}/{Sector}{year}{Quantity}{suffix}"
                )
            )
            overwrite = params.pop("overwrite", False)
            sector = params.pop("sector", "")
            lsuffix = params.pop("suffix", suffix)
            if lsuffix is not None and len(lsuffix) > 0 and lsuffix[0] != ".":
                lsuffix = "." + lsuffix
            # assumes directory if filename has no suffix and does not exist
            if getattr(quantity, "name", None) is None:
                name = "quantity"
            else:
                name = str(quantity.name)

            filename = Path(
                filestring.format(
                    cwd=str(Path().absolute()),
                    quantity=name,
                    Quantity=name.title(),
                    sector=sector,
                    Sector=sector.title(),
                    year=year,
                    suffix=lsuffix,
                    default_output_dir=str(DEFAULT_OUTPUT_DIRECTORY),
                )
            )
            if filename.exists():
                if overwrite:
                    filename.unlink()
                else:
                    msg = (
                        f"File {filename} already exists and overwrite argument has "
                        "not been given."
                    )
                    getLogger(function.__module__).critical(msg)
                    raise OSError(msg)

            filename.parent.mkdir(parents=True, exist_ok=True)
            function(quantity, filename, **params)  # type: ignore
            return filename

        return decorated

    return decorator


def standardize_quantity(
    function: Callable[[pd.DataFrame | xr.DataArray, str], None],
):
    """Helps standardize how the quantities are specified.

    This decorator adds three keyword arguments to an input function:

    - set_index: A dictionary or any argument accepted by
      :py:meth:`pandas.DataFrame.set_index`. Ignored if not specified. If specified, a
      call to :py:meth:`pandas.DataFrame.reset_index` is made first.
    - sort_index: A dictionary or any argument accepted by
      :py:meth:`pandas.DataFrame.sort_index`. Ignored if not specified.
    - keep_columns: a string or a list of strings with the names of the columns to keep.
      Ignored if not specified.

    The three functions are applied in the order given, assuming an input is specified.
    """
    from functools import wraps

    class NotSpecified:
        pass

    @wraps(function)
    def decorated(
        quantity: pd.DataFrame | xr.DataArray,
        *args,
        set_index: Any | NotSpecified = NotSpecified,
        sort_index: Any | NotSpecified | bool = NotSpecified,
        keep_columns: str | Sequence[str] | NotSpecified = NotSpecified,
        group_by: Any | NotSpecified = NotSpecified,
        **config,
    ) -> None:
        any_calls = (
            set_index is not NotSpecified
            or set_index is not NotSpecified
            or keep_columns is not None
        )
        if any_calls and hasattr(quantity, "to_dataframe"):
            data: pd.DataFrame = quantity.to_dataframe()
        else:
            data = cast(pd.DataFrame, quantity)
        if isinstance(set_index, Mapping):
            data = data.reset_index().set_index(**set_index)
        elif set_index is not NotSpecified:
            data = data.reset_index().set_index(set_index)
        if isinstance(sort_index, Mapping):
            data = data.sort_index(**sort_index)
        elif sort_index is True:
            data = data.sort_index()
        elif sort_index is not NotSpecified and sort_index is not False:
            data = data.sort_index(sort_index)
        if keep_columns is not NotSpecified:
            data = data[keep_columns]
        if isinstance(group_by, Mapping):
            data = data.groupby(group_by).sum()
        elif group_by is not NotSpecified:
            data = data.groupby(group_by).sum()

        return function(data, *args, **config)

    return decorated


@register_output_sink(name="csv")
@sink_to_file(".csv")
@standardize_quantity
def to_csv(quantity: pd.DataFrame | xr.DataArray, filename: str, **params) -> None:
    """Saves data array to csv format, using pandas.to_csv.

    Arguments:
        quantity: The data to be saved
        filename: File to which the data should be saved
        params: A configuration dictionary accepting any argument to `pandas.to_csv`
    """
    if "float_format" not in params:
        params["float_format"] = "%.4f"
    if "index" not in params:
        params["index"] = False
    if "columns" in params:
        quantity = quantity.reset_index()
    if isinstance(quantity, xr.DataArray):
        quantity = quantity.to_dataframe()
    quantity.to_csv(filename, **params)


@register_output_sink(name=("netcdf", "nc"))
@sink_to_file(".nc")
def to_netcdf(quantity: xr.DataArray | pd.DataFrame, filename: str, **params) -> None:
    """Saves data array to csv format, using xarray.to_netcdf.

    Arguments:
        quantity: The data to be saved
        filename: File to which the data should be saved
        params: A configuration dictionary accepting any argument to `xarray.to_netcdf`
    """
    name = quantity.name if getattr(quantity, "name", None) is not None else "quantity"
    assert name is not None
    if isinstance(quantity, pd.DataFrame):
        dataset = xr.Dataset.from_dataframe(quantity)
    else:
        dataset = xr.Dataset({name: quantity})
    dataset.to_netcdf(filename, **params)


@register_output_sink(name=("excel", "xlsx"))
@sink_to_file(".xlsx")
@standardize_quantity
def to_excel(quantity: pd.DataFrame | xr.DataArray, filename: str, **params) -> None:
    """Saves data array to csv format, using pandas.to_excel.

    Arguments:
        quantity: The data to be saved
        filename: File to which the data should be saved
        params: A configuration dictionary accepting any argument to `pandas.to_excel`
    """
    from logging import getLogger

    if isinstance(quantity, xr.DataArray):
        quantity = quantity.to_dataframe()
    try:
        quantity.to_excel(filename, **params)
    except ModuleNotFoundError as e:
        msg = f"Cannot save to excel format: missing python package ({e})"
        getLogger(__name__).critical(msg)
        raise


@register_output_sink(name="aggregate")
class YearlyAggregate:
    """Incrementally aggregates data from year to year."""

    def __init__(
        self,
        final_sink: MutableMapping[str, Any] | None = None,
        sector: str = "",
        axis="year",
        **kwargs,
    ):
        if final_sink is None:
            final_sink = dict(**kwargs)
        else:
            final_sink["sink"].update(**kwargs)
        if "overwrite" not in final_sink and not (
            isinstance(final_sink.get("sink", None), Mapping)
            and "overwrite" in final_sink["sink"]
        ):
            final_sink["overwrite"] = True
        self.sink = factory(final_sink, sector_name=sector)
        self.aggregate: pd.Dataframe | None = None
        self.axis = axis

    def __call__(self, data: pd.DataFrame | xr.DataArray, year: int):
        if isinstance(data, xr.DataArray):
            dataframe = data.to_dataframe()
        else:
            dataframe = data
        if self.axis in dataframe.columns:
            dataframe = dataframe[dataframe[self.axis] == year]
        elif self.axis in getattr(dataframe.index, "names", []):
            dataframe = dataframe.xs(year, level=self.axis).assign(year=year)
        if self.aggregate is None:
            self.aggregate = dataframe
        else:
            self.aggregate = pd.concat((self.aggregate, dataframe), sort=True)
        assert self.aggregate is not None
        if getattr(data, "name", None) is not None:
            self.aggregate.name = data.name
        return self.sink(self.aggregate, year=year)


class FiniteResourceException(Exception):
    """Raised when a finite resource is exceeded."""


@register_output_sink
def finite_resource_logger(
    data: pd.DataFrame | xr.DataArray, year: int, early_exit=False, **kwargs
):
    from logging import getLogger

    if data is None:
        return

    over_limit = (~data).any([u for u in data.dims if u != "commodity"])
    over_limit = over_limit.sel(commodity=over_limit)
    if over_limit.size != 0:
        msg = "The following commodities have exceeded their limits: " + ", ".join(
            [str(u) for u in over_limit.commodity.values]
        )
        getLogger(__name__).critical(msg)
        if early_exit:
            raise FiniteResourceException(msg)



"""Ensemble of functions to read MUSE data."""

from __future__ import annotations

__all__ = ["read_settings"]

import importlib.util as implib
from collections import namedtuple
from collections.abc import Mapping, MutableMapping, Sequence
from copy import deepcopy
from logging import getLogger
from pathlib import Path
from typing import Any, Callable

import numpy as np
import xarray as xr

from muse.defaults import DATA_DIRECTORY

DEFAULT_SETTINGS_PATH = DATA_DIRECTORY / "default_settings.toml"
"""Default settings path."""

SETTINGS_HOOKS_SIGNATURE = Callable[[dict], None]
"""settings checks signature."""

SETTINGS_HOOKS: list[tuple[int, str, SETTINGS_HOOKS_SIGNATURE]] = []
"""Dictionary of settings checks."""


class InputError(Exception):
    """Root for TOML input errors."""


class MissingSettings(InputError):
    """Error when an input is missing."""


class IncorrectSettings(InputError):
    """Error when an input exists but is incorrect."""


def convert(dictionary: dict) -> namedtuple:
    """Converts a dictionary (with nested ones) to a nametuple."""
    for key, value in dictionary.items():
        if isinstance(value, dict):
            dictionary[key] = convert(value)
    return namedtuple("MUSEOptions", dictionary.keys())(**dictionary)


def undo_damage(nt) -> Any:
    """Unconvert nested nametuple."""
    if not hasattr(nt, "_asdict"):
        return nt
    result = nt._asdict()
    for key, value in result.items():
        result[key] = undo_damage(value)
    return result


class FormatDict(dict):
    """Allows partial formatting of a string."""

    def __missing__(self, key):
        return FormatDict.FormatPlaceholder(key)

    class FormatPlaceholder:
        def __init__(self, key):
            self.key = key

        def __format__(self, spec):
            result = f"{self.key}:{spec}" if spec else self.key
            return f"{{{result}}}"


def format_path(
    filepath: str,
    replacements: Mapping | None = None,
    path: str | Path | None = None,
    cwd: str | Path | None = None,
) -> Path:
    """Replaces known patterns in a path.

    Unknown patterns are left alone. This allows downstream object factories to format
    the paths according to their own specifications.
    """
    from string import Formatter

    patterns = FormatDict(
        {
            **{
                "cwd": Path("" if cwd is None else cwd).absolute(),
                "path": Path("" if path is None else path).absolute(),
            },
            **({} if replacements is None else replacements),
        }
    )
    formatter = Formatter()
    return Path(formatter.vformat(str(filepath), (), patterns)).absolute()


def format_paths(
    settings: Mapping,
    path: Path,
    cwd: Path,
    suffixes: Sequence[str] = (".csv", ".nc", ".xls", ".xlsx", ".py", ".toml"),
) -> dict:
    """Format paths passed to settings.

    This function is used to format paths in the settings file. It is used to replace
    the {path} and {cwd} placeholders with the actual path and current working
    directory.

    Args:
        settings: The settings dictionary to format
        path: The path to the settings file
        cwd: The current working directory
        suffixes: Suffixes used to identify strings as paths
    """

    def is_a_path(key, value):
        return (
            isinstance(value, (str, Path)) and Path(value).suffix in suffixes
        ) or key == "filename"

    # Recursively format paths
    result = dict(**settings)
    for key, value in result.items():
        if is_a_path(key, value):
            result[key] = format_path(value, path=path, cwd=cwd)
        elif isinstance(value, Mapping):
            result[key] = format_paths(settings=value, path=path, cwd=cwd)
        elif isinstance(value, list):
            result[key] = [
                format_paths(settings=item, path=path, cwd=cwd)
                if isinstance(item, Mapping)
                else format_path(item, path=path, cwd=cwd)
                if is_a_path(key, item)
                else item
                for item in result[key]
            ]

    return result


def read_toml(tomlfile: str | Path, path: str | Path | None = None) -> MutableMapping:
    """Reads a TOML file and formats the paths.

    Args:
        tomlfile: Path to the TOML file (string or Path object)
        path: Optional path to use for formatting relative paths (string or Path object)

    Returns:
        MutableMapping containing the formatted TOML data
    """
    from toml import load

    tomlfile = Path(tomlfile)
    toml = load(tomlfile)
    if path is None:
        path = tomlfile.parent
    else:
        path = Path(path)
    settings = format_paths(toml, path=path, cwd=Path())
    return settings


def read_settings(settings_file: str | Path) -> namedtuple:
    """Loads the input settings for any MUSE simulation.

    Loads a MUSE settings file. This must be a TOML formatted file. Missing settings are
    loaded from the DEFAULT_SETTINGS. Custom Python modules, if present, are loaded
    and hooks are run to process and validate the settings and ensure that they are
    compatible with a MUSE simulation.

    Arguments:
        settings_file: A string or a Path to the settings file

    Returns:
        A dictionary with the settings
    """
    getLogger(__name__).info("Reading MUSE settings")
    settings_file = Path(settings_file)

    # The user data
    user_settings = read_toml(settings_file)

    # Get default settings
    default_path = Path(user_settings.get("default_settings", DEFAULT_SETTINGS_PATH))
    default_settings = read_toml(default_path, path=settings_file.parent)

    # Timeslice information cannot be merged. Accept only information from one.
    if "timeslices" in user_settings:
        default_settings.pop("timeslices", None)

    # We update the default information with the user provided data
    settings = add_known_parameters(default_settings, user_settings)
    settings = add_unknown_parameters(settings, user_settings)

    # Finally, we run some hooks to make sure all makes sense and files exist.
    process_settings(settings)
    return convert(settings)


def add_known_parameters(default_dict, user_dict, parent=None) -> dict:
    """Recursively merge user settings with default settings.

    Validates required parameters and handles optional ones.

    Args:
        default_dict: Dictionary containing default settings
        user_dict: Dictionary containing user-provided settings
        parent: Parent key for nested dictionaries (used for logging)

    Returns:
        Merged dictionary with validated settings
    """
    from logging import getLogger

    merged = deepcopy(default_dict)
    defaults_used = []
    missing = []

    for key in default_dict:
        if key in user_dict:
            value = user_dict[key]
            if isinstance(value, Mapping):
                new_parent = f"{parent}.{key}" if parent else key
                merged[key] = add_known_parameters(
                    merged.get(key, {}), value, new_parent
                )
            else:
                merged[key] = value
        elif isinstance(merged[key], str):
            if merged[key].lower() == "required":
                missing.append(key)
            elif merged[key].lower() == "optional":
                merged.pop(key)
        else:
            defaults_used.append(f"{parent}.{key}" if parent else key)

    if missing:
        raise MissingSettings(f"Required parameters missing in input file: {missing}")

    if defaults_used:
        getLogger(__name__).info(
            f"Default input values used: {', '.join(defaults_used)}"
        )

    return merged


def add_unknown_parameters(default_dict, user_dict) -> dict:
    """Recursively merge user settings with default settings.

    Preserves unknown parameters from user settings.

    Args:
        default_dict: Dictionary containing default settings
        user_dict: Dictionary containing user-provided settings

    Returns:
        Merged dictionary containing both default and user settings
    """
    merged = deepcopy(default_dict)

    for key, value in user_dict.items():
        if isinstance(value, Mapping):
            merged[key] = add_unknown_parameters(merged.get(key, {}), value)
        else:
            merged[key] = value

    return merged


def process_settings(settings: dict) -> None:
    """Run the hooks on the settings file."""
    msg = " Processing input settings..."
    getLogger(__name__).info(msg)

    # Load extra hooks from plugins
    check_plugins(settings)
    # This must be run before the other hooks to ensure that custom defined settings
    # hooks are all loaded before validating the settings.

    # Run hooks in order of priority
    for _, _, hook in sorted(SETTINGS_HOOKS, key=lambda x: x[0]):
        hook(settings)


def check_plugins(settings: dict) -> None:
    """Check and load user-defined Python plugin files if they exist."""
    plugins = settings.get("plugins", [])

    # Handle plugins as dict, str, or Path
    if isinstance(plugins, (dict, Mapping)):
        plugins = plugins.get("plugins", [])
    if isinstance(plugins, (Path, str)):
        plugins = [plugins]
    if not plugins:
        return

    for plugin in plugins:
        plugin_path = Path(format_path(plugin))
        if not plugin_path.exists():
            msg = f"ERROR plugin does not exist: {plugin_path}"
            getLogger(__name__).critical(msg)
            raise IncorrectSettings(msg)

        # Load the plugin module
        spec = implib.spec_from_file_location(plugin_path.stem, plugin_path)
        mod = implib.module_from_spec(spec)
        spec.loader.exec_module(mod)  # type: ignore
        getLogger(__name__).info(f"Loaded plugin {plugin_path.stem} from {plugin_path}")


def register_settings_hook(
    func: SETTINGS_HOOKS_SIGNATURE | None = None, *, priority: int = 100
) -> Callable:
    """Register a function to be called during settings validation.

    The function will be called with the settings dictionary as its only argument.
    The function can modify the settings dictionary in place.

    Args:
        func: The function to register
        priority: The priority of the function. Lower numbers are called first.

    Returns:
        The decorated function
    """

    def decorated(f: SETTINGS_HOOKS_SIGNATURE) -> SETTINGS_HOOKS_SIGNATURE:
        """Register the function and return it unchanged."""
        getLogger(__name__).debug(
            f"Registering settings hook {f.__name__} with priority {priority}"
        )
        SETTINGS_HOOKS.append((priority, f.__name__, f))
        return f

    if func is None:
        return decorated
    return decorated(func)


@register_settings_hook(priority=0)
def standardise_case(settings: dict) -> None:
    """Standardise certain fields to snake_case."""
    from muse.utilities import camel_to_snake

    fields_to_standardise = ["excluded_commodities", "regions"]
    for field in fields_to_standardise:
        if field in settings:
            settings[field] = [camel_to_snake(x) for x in settings[field]]

    # Handle timeslice level_names if present
    if "level_names" in settings["timeslices"]:
        settings["timeslices"]["level_names"] = [
            camel_to_snake(x) for x in settings["timeslices"]["level_names"]
        ]


@register_settings_hook(priority=1)
def check_sectors(settings: dict) -> None:
    """Check that there is at least 1 sector."""
    assert len(settings["sectors"]) >= 1, "ERROR - There must be at least 1 sector."


@register_settings_hook(priority=1)
def setup_timeslices(settings: dict) -> None:
    """Set up the timeslices."""
    from muse.timeslices import setup_module

    setup_module(settings)
    settings.pop("timeslices", None)


@register_settings_hook(priority=1)
def setup_commodities(settings: dict) -> None:
    """Set up the commodities."""
    from muse.commodities import setup_module

    setup_module(settings["global_input_files"]["global_commodities"])


@register_settings_hook(priority=1)
def setup_time_framework(settings: dict) -> None:
    """Converts the time framework to a sorted array."""
    settings["time_framework"] = np.array(sorted(settings["time_framework"]), dtype=int)


@register_settings_hook
def check_log_level(settings: dict) -> None:
    """Check the log level required in the simulation."""
    valid_levels = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
    msg = "ERROR - Valid log levels are {}.".format(", ".join(valid_levels))
    assert settings["log_level"].upper() in valid_levels, msg

    settings["log_level"] = settings["log_level"].upper()


@register_settings_hook
def check_interpolation_mode(settings: dict) -> None:
    """Checks that the interpolation mode is valid."""
    settings["interpolation_mode"] = settings["interpolation_mode"].lower()

    valid_modes = [
        "linear",
        "nearest",
        "zero",
        "slinear",
        "quadratic",
        "cubic",
        "previous",
        "next",
        "active",  # legacy: see below
    ]
    msg = (
        'ERROR - Valid interpolation modes are "linear", "nearest", "zero", '
        '"slinear", "quadratic", "cubic", "previous", "next"'
    )
    assert settings["interpolation_mode"] in valid_modes, msg

    # Legacy: "Active" was previous default - switch to "linear" (#642)
    if settings["interpolation_mode"] == "active":
        msg = "'Active' interpolation mode is deprecated. Defaulting to 'linear'."
        getLogger(__name__).warning(msg)
        settings["interpolation_mode"] = "linear"


@register_settings_hook
def check_budget_parameters(settings: dict) -> None:
    """Check the parameters that are required if carbon_budget > 0."""
    budget = settings["carbon_budget_control"]["budget"]
    time_framework = settings["time_framework"]
    if not budget:
        settings["carbon_budget_control"]["budget"] = xr.DataArray([])
        return

    msg = "ERROR - budget_check must have the same length that time_framework"
    if len(budget) != len(time_framework):
        raise AssertionError(msg)

    coords = time_framework
    settings["carbon_budget_control"]["budget"] = xr.DataArray(
        np.array(budget), dims="year", coords={"year": coords}
    )


@register_settings_hook
def check_iteration_control(settings: dict) -> None:
    """Check and set iteration control parameters for equilibrium and convergence."""
    equilibrium = str(settings["equilibrium"]).lower()
    if equilibrium in ("false", "off"):
        settings["equilibrium"] = False
        return

    settings["equilibrium"] = True
    if settings["maximum_iterations"] <= 0:
        raise ValueError("ERROR - The number of iterations must be a positive number.")
    settings["maximum_iterations"] = int(settings["maximum_iterations"])
    if settings["tolerance"] <= 0:
        raise ValueError("ERROR - The convergence tolerance must be a positive number.")


@register_settings_hook
def sort_sectors(settings: dict) -> None:
    """Set the priorities of the sectors."""
    sectors = settings["sectors"]
    priorities = {
        "preset": 0,
        "presets": 0,
        "demand": 10,
        "conversion": 20,
        "supply": 30,
        "last": 100,
    }

    # If sectors has a 'list' key, flatten it
    if "list" in sectors:
        sectors = {k: sectors[k] for k in sectors["list"]}

    for sector in sectors.values():
        # Assign priority, using default if not present or not recognized
        prio = sector.get("priority", priorities["last"])
        sector["priority"] = int(priorities.get(str(prio).lower().strip(), prio))

    # Sort sector names by priority
    sectors["list"] = sorted(sectors.keys(), key=lambda x: sectors[x]["priority"])
    settings["sectors"] = sectors


@register_settings_hook
def check_currency(settings: dict) -> None:
    """Raise a warning if no currency is specified."""
    if not settings.get("currency", None):
        msg = (
            "No currency specified. Please specify a currency in the settings file "
            "using the 'currency' parameter."
        )
        getLogger(__name__).warning(msg)
        return


@register_settings_hook
def check_deprecated_params(settings: dict) -> None:
    """Check for and warn about deprecated parameters."""
    deprecated_params = ["foresight", "interest_rate"]
    for param in deprecated_params:
        if param in settings:
            msg = (
                f"The `{param}` parameter has been deprecated. "
                "Please remove it from your settings file."
            )
            getLogger(__name__).warning(msg)
            settings.pop(param)


@register_settings_hook(priority=10)
def check_subsector_settings(settings: dict) -> None:
    """Check for invalid or deprecated subsector settings.

    Validates:
    - Renamed asset_threshhold parameter (PR #447)
    - Missing lpsolver parameter (PR #587)
    - Deprecated forecast parameter (PR #645)
    """
    from logging import getLogger

    # Check each sector's subsectors
    for sector_name, sector in settings["sectors"].items():
        if "subsectors" not in sector:
            continue

        for subsector_name, subsector in sector["subsectors"].items():
            # Check for renamed asset_threshhold parameter
            if "asset_threshhold" in subsector:
                msg = (
                    "Invalid parameter asset_threshhold. Did you mean asset_threshold?"
                )
                raise ValueError(msg)

            # Check for missing lpsolver
            if "lpsolver" not in subsector:
                msg = (
                    f"lpsolver not specified for subsector '{subsector_name}' "
                    f"in sector '{sector_name}'. Defaulting to 'scipy'"
                )
                getLogger(__name__).warning(msg)

            # Check for deprecated forecast parameter
            if "forecast" in subsector:
                msg = (
                    "The 'forecast' parameter has been deprecated. "
                    "Please remove from your settings file."
                )
                getLogger(__name__).warning(msg)


def read_technodata(
    settings: Any,
    sector_name: str,
    interpolation_mode: str = "linear",
) -> xr.Dataset:
    """Read and process technodata for a given sector.

    This function reads technology data from CSV files and processes it for use in MUSE
    simulations. It handles technology specifications, trade data, and interpolates
    the data to match the simulation timeframe.

    Args:
        settings: MUSE settings object containing configuration parameters
        sector_name: Name of the sector to read technodata for
        interpolation_mode: Method for interpolating data between years.
            Defaults to "linear"

    Returns:
        xr.Dataset: Processed technodata containing technology specifications,
            inputs/outputs, and trade information
    """
    from muse.readers.csv import read_technologies, read_trade_technodata

    regions = settings.regions
    time_framework = settings.time_framework
    settings = getattr(settings.sectors, sector_name)

    # Legacy: technodata settings could be in a "technodata" section
    if isinstance(undo_damage(settings.technodata), Mapping):
        settings = settings.technodata

    # Read technodata
    technologies = read_technologies(
        technodata_path=Path(settings.technodata),
        technodata_timeslices_path=getattr(settings, "technodata_timeslices", None),
        comm_out_path=Path(settings.commodities_out),
        comm_in_path=Path(settings.commodities_in),
        time_framework=time_framework,
        interpolation_mode=interpolation_mode,
    ).sel(region=regions)

    # Only keep commodities that are used as inputs or outputs
    dims = ("year", "region", "technology")
    fixed_ins = (technologies.fixed_inputs > 0).any(
        [d for d in dims if d in technologies.fixed_inputs.dims]
    )
    flex_ins = (technologies.flexible_inputs > 0).any(
        [d for d in dims if d in technologies.flexible_inputs.dims]
    )
    outs = (technologies.fixed_outputs > 0).any(
        [d for d in dims if d in technologies.fixed_outputs.dims]
    )
    techcomms = technologies.commodity[fixed_ins | flex_ins | outs]
    technologies = technologies.sel(commodity=techcomms)

    # Read trade technodata
    if hasattr(settings, "trade"):
        trade_data = read_trade_technodata(settings.trade)
        if "region" in trade_data.dims:
            trade_data = trade_data.sel(region=regions)
        if "dst_region" in trade_data.dims:
            trade_data = trade_data.sel(dst_region=regions)
            if trade_data.dst_region.size == 1:
                trade_data = trade_data.squeeze("dst_region", drop=True)

        # Drop duplicate data vars before merging
        common_vars = set(technologies.data_vars) & set(trade_data.data_vars)
        technologies = technologies.drop_vars(common_vars)
        technologies = technologies.merge(trade_data)

    technologies = technologies.set_index(commodity="commodity")  # See PR #638
    return technologies


def read_presets_sector(settings: Any, sector_name: str) -> xr.Dataset:
    """Read data for a preset sector.

    This function reads consumption and supply data for a preset sector from various
    data sources. It supports multiple input formats including direct consumption data,
    demand tables, or correlation-based consumption calculated from macro drivers and
    regression parameters.

    Args:
        settings: MUSE settings object containing configuration parameters
        sector_name: Name of the preset sector to read data for

    Returns:
        xr.Dataset: Dataset containing consumption and supply data for the sector.
            Costs are initialized to zero.
    """
    from muse.readers import read_attribute_table, read_presets
    from muse.timeslices import distribute_timeslice, drop_timeslice

    sector_conf = getattr(settings.sectors, sector_name)

    # Read consumption data
    if getattr(sector_conf, "consumption_path", None) is not None:
        consumption = read_presets(sector_conf.consumption_path)
    elif getattr(sector_conf, "demand_path", None) is not None:
        consumption = read_attribute_table(sector_conf.demand_path)
        if "timeslice" not in consumption.dims:
            consumption = distribute_timeslice(consumption)
    elif (
        getattr(sector_conf, "macrodrivers_path", None) is not None
        and getattr(sector_conf, "regression_path", None) is not None
    ):
        consumption = read_correlation_consumption(sector_conf)
    else:
        raise MissingSettings(f"Missing consumption data for sector {sector_name}")

    # Create presets dataset
    presets = xr.Dataset(
        {
            "consumption": consumption,
            "supply": read_presets(sector_conf.supply_path)
            if getattr(sector_conf, "supply_path", None) is not None
            else drop_timeslice(xr.zeros_like(consumption)),
            "costs": drop_timeslice(xr.zeros_like(consumption)),
        }
    )

    return presets


def read_correlation_consumption(sector_conf: Any) -> xr.Dataset:
    """Read consumption data for a sector based on correlation files.

    This function calculates endogenous demand for a sector using macro drivers and
    regression parameters. It applies optional filters, handles sector aggregation,
    and distributes the consumption across timeslices if timeslice shares are provided.

    Args:
        sector_conf: Sector configuration object containing paths to macro drivers,
            regression parameters, and timeslice shares files

    Returns:
        xr.Dataset: Consumption data distributed across timeslices and regions
    """
    from muse.readers import (
        read_macro_drivers,
        read_regression_parameters,
        read_timeslice_shares,
    )
    from muse.regressions import endogenous_demand
    from muse.timeslices import broadcast_timeslice, distribute_timeslice

    macro_drivers = read_macro_drivers(sector_conf.macrodrivers_path)
    regression_parameters = read_regression_parameters(sector_conf.regression_path)
    consumption = endogenous_demand(
        drivers=macro_drivers,
        regression_parameters=regression_parameters,
        forecast=0,
    )

    # Legacy: apply filters
    if hasattr(sector_conf, "filters"):
        consumption = consumption.sel(sector_conf.filters._asdict())

    # Legacy: we permit regression parameters to split by sector, so have to sum
    if "sector" in consumption.dims:
        consumption = consumption.sum("sector")

    # Split by timeslice
    if sector_conf.timeslice_shares_path is not None:
        shares = read_timeslice_shares(sector_conf.timeslice_shares_path)
        consumption = broadcast_timeslice(consumption) * shares
    else:
        consumption = distribute_timeslice(consumption)

    return consumption



"""Test saving outputs to file."""

from pathlib import Path
from unittest.mock import patch

import numpy as np
import pandas as pd
import xarray as xr
from pytest import approx, fixture, importorskip, mark, raises

from muse.outputs.sector import factory


@fixture
def streetcred(save_registries):
    """Create a test output quantity that returns random data."""
    from muse.outputs.sector import register_output_quantity

    @register_output_quantity
    def streetcred(*args, **kwargs):
        return xr.DataArray(
            np.random.randint(0, 5, (3, 2)),
            coords={
                "year": [2010, 2015],
                "technology": ("asset", ["a", "b", "c"]),
                "installed": ("asset", [2010, 2011, 2011]),
            },
            dims=("asset", "year"),
        )


@fixture
def market():
    """Common market fixture used in multiple tests."""
    return xr.DataArray([1], coords={"year": [2010]}, dims="year")


@fixture
def base_config(tmp_path):
    """Common config fixture used in multiple tests."""
    path = tmp_path / "results" / "stuff"
    return {
        "filename": path / "{Sector}{year}{Quantity}.csv",
        "quantity": "streetcred",
    }


def create_test_data_array(values, coords=None, name="test"):
    """Helper function to create test DataArrays."""
    if coords is None:
        coords = dict(a=[2, 4])
    return xr.DataArray(values, coords=coords, dims="a", name=name)


def assert_file_exists_and_readable(path, expected_columns=None):
    """Helper to verify file exists and can be read."""
    assert path.exists() and path.is_file()
    df = pd.read_csv(path)
    if expected_columns:
        assert set(df.columns) == set(expected_columns)
    return df


@mark.usefixtures("streetcred")
def test_save_with_dir(tmp_path, market, base_config):
    """Test saving output to directory with sector and year in filename."""
    result = factory(base_config, sector_name="Yoyo")(market, None, None)
    assert len(result) == 1
    expected_path = Path(base_config["filename"]).parent / "Yoyo2010Streetcred.csv"
    assert result[0] == expected_path
    assert_file_exists_and_readable(result[0])


@mark.usefixtures("streetcred")
def test_overwrite(tmp_path, market, base_config):
    """Test file overwrite behavior."""
    outputter = factory(base_config, sector_name="Yoyo")
    result = outputter(market, None, None)
    expected_path = Path(base_config["filename"]).parent / "Yoyo2010Streetcred.csv"
    assert result[0] == expected_path
    assert_file_exists_and_readable(result[0])

    # default is to never overwrite
    with raises(IOError):
        outputter(market, None, None)

    base_config["overwrite"] = True
    factory(base_config, sector_name="Yoyo")(market, None, None)


@mark.usefixtures("streetcred")
@mark.parametrize(
    "config_type,suffix",
    [
        ("suffix", "nc"),
        ("sink", "nc"),
    ],
)
def test_save_with_path_to_nc(tmp_path, market, base_config, config_type, suffix):
    """Test saving output to NC file with different config types."""
    path = tmp_path / "results" / "stuff"
    if config_type == "suffix":
        config = {
            "filename": path / "{Sector}{year}{Quantity}{suffix}",
            "quantity": "streetcred",
            "suffix": suffix,
        }
    else:
        config = {
            "filename": path / "{sector}{year}{quantity}.csv",
            "quantity": "streetcred",
            "sink": suffix,
        }
    result = factory(config, sector_name="Yoyo")(market, None, None)
    expected_path = path / (
        f"{'Yoyo' if config_type == 'suffix' else 'yoyo'}"
        f"2010"
        f"{'Streetcred' if config_type == 'suffix' else 'streetcred'}"
        f".{'nc' if config_type == 'suffix' else 'csv'}"
    )
    assert result[0] == expected_path
    assert result[0].is_file()
    xr.open_dataset(result[0])


@mark.usefixtures("streetcred")
def test_save_with_fullpath_to_excel(tmp_path):
    from warnings import simplefilter

    from pandas import read_excel

    importorskip("openpyxl")
    simplefilter("default", PendingDeprecationWarning)

    path = tmp_path / "results" / "stuff" / "this.xlsx"
    config = {"filename": path, "quantity": "streetcred", "sink": "xlsx"}
    market = xr.DataArray([1], coords={"year": [2010]}, dims="year")
    # can use None because we **know** none of the arguments are used here
    result = factory(config, sector_name="Yoyo")(market, None, None)
    assert result[0] == path
    assert result[0].is_file()
    read_excel(result[0])


@patch("muse.outputs.cache.consolidate_quantity")
def test_output_functions(mock_consolidate):
    """Test output functions (capacity, production, lcoe) with common setup."""
    from muse.outputs.cache import capacity, lcoe, production

    cached = [xr.DataArray() for _ in range(3)]
    agents = {}

    for func, quantity in [
        (capacity, "capacity"),
        (production, "production"),
        (lcoe, "lcoe"),
    ]:
        func(cached, agents)
        mock_consolidate.assert_called_once_with(quantity, cached, agents)
        mock_consolidate.reset_mock()


@mark.usefixtures("streetcred")
def test_no_sink_or_suffix(tmp_path, market):
    """Test default sink and suffix behavior."""
    config = dict(
        quantity="streetcred",
        filename=f"{tmp_path}/{{Sector}}{{Quantity}}{{year}}{{suffix}}",
    )
    result = factory(config)(market, None, None)
    assert len(result) == 1
    assert result[0].is_file()
    assert result[0].suffix == ".csv"


@mark.usefixtures("save_registries")
def test_can_register_class():
    """Test class registration functionality."""
    from muse.outputs.sinks import factory, register_output_sink

    @register_output_sink
    class AClass:
        def __init__(self, sector, some_args=3):
            self.sector = sector
            self.some_args = some_args

        def __call__(self, x):
            pass

    # Test default arguments
    settings = {"sink": {"name": "AClass"}}
    sink = factory(settings, sector_name="yoyo")
    assert isinstance(sink, AClass)
    assert sink.sector == "yoyo"
    assert sink.some_args == 3

    # Test custom arguments
    settings = {"sink": {"name": "AClass", "some_args": 5}}
    sink = factory(settings, sector_name="yoyo")
    assert isinstance(sink, AClass)
    assert sink.sector == "yoyo"
    assert sink.some_args == 5


@mark.usefixtures("save_registries")
def test_can_register_function():
    """Test function registration functionality."""
    from muse.outputs.sinks import factory, register_output_sink

    @register_output_sink
    def a_function(*args):
        pass

    settings = {"sink": "a_function"}
    sink = factory(settings, sector_name="yoyo")
    assert sink.func is a_function


@mark.usefixtures("save_registries")
def test_yearly_aggregate():
    """Test yearly aggregation with custom sink."""
    from muse.outputs.sinks import factory, register_output_sink

    # Setup tracking variables
    received_data = None
    gyear = None
    gsector = None
    goverwrite = None

    class MySpecialReturn:
        pass

    @register_output_sink(overwrite=True)
    def dummy(data, year: int, sector: str, overwrite: bool) -> MySpecialReturn:
        nonlocal received_data, gyear, gsector, goverwrite
        received_data = data
        gyear = year
        gsector = sector
        goverwrite = overwrite
        return MySpecialReturn()

    sink = factory(
        dict(overwrite=True, sink=dict(aggregate="dummy")), sector_name="yoyo"
    )

    # Test first year
    data = create_test_data_array([1, 0], name="nada")
    data["year"] = 2010
    assert isinstance(sink(data, 2010), MySpecialReturn)
    assert gyear == 2010
    assert gsector == "yoyo"
    assert goverwrite is True
    assert isinstance(received_data, pd.DataFrame)

    # Test second year
    data = create_test_data_array([0, 1], name="nada")
    data["year"] = 2020
    assert isinstance(sink(data, 2020), MySpecialReturn)
    assert gyear == 2020
    assert gsector == "yoyo"
    assert received_data[received_data.year == 2010].nada.values == approx([1, 0])
    assert received_data[received_data.year == 2020].nada.values == approx([0, 1])


def test_yearly_aggregate_file(tmp_path):
    """Test yearly aggregation to file with multiple years of data."""
    from muse.outputs.sinks import factory

    path = tmp_path / "file.csv"
    sink = factory(dict(filename=str(path), sink="aggregate"), sector_name="yoyo")

    def verify_year_data(values, year, expected_rows):
        data = create_test_data_array(values, name="georges")
        data["year"] = year
        assert sink(data, year) == path
        df = assert_file_exists_and_readable(path, {"year", "georges"})
        assert df.shape[0] == expected_rows
        return df

    # Test first year
    verify_year_data([1, 0], 2010, 2)

    # Test second year (should append to existing file)
    df2 = verify_year_data([0, 1], 2020, 4)

    # Verify data from both years is present
    assert set(df2.year.unique()) == {2010, 2020}
    assert df2[df2.year == 2010].georges.tolist() == [1, 0]
    assert df2[df2.year == 2020].georges.tolist() == [0, 1]


def test_yearly_aggregate_no_outputs(tmp_path):
    """Test behavior with no outputs configured."""
    from muse.outputs.mca import factory

    outputs = factory()
    assert len(outputs(None, year=2010)) == 0


def setup_mca_test(tmp_path, outputs_config):
    """Helper function to set up MCA tests."""
    from toml import dumps, load

    from muse import examples
    from muse.mca import MCA

    examples.copy_model(path=tmp_path)
    settings = load(tmp_path / "model" / "settings.toml")
    settings["outputs"] = [outputs_config]
    settings["time_framework"] = settings["time_framework"][:2]
    file = tmp_path / "model" / "settings.toml"
    file.write_text(dumps(settings), encoding="utf-8")
    return MCA.factory(file)


def test_mca_aggregate_outputs(tmp_path):
    """Test MCA aggregate outputs."""
    mca = setup_mca_test(
        tmp_path,
        dict(filename="{path}/{Quantity}{suffix}", quantity="prices", sink="aggregate"),
    )
    mca.run()
    assert (tmp_path / "model" / "Prices.csv").exists()
    # TODO: should pass again after #612
    # data = pd.read_csv(tmp_path / "model" / "Prices.csv")
    # assert set(data.year) == set(settings["time_framework"])


@mark.usefixtures("save_registries")
def test_path_formatting(tmp_path):
    """Test path formatting with dummy sink and quantity."""
    from muse.outputs.mca import register_output_quantity
    from muse.outputs.sinks import register_output_sink, sink_to_file

    @register_output_sink(name="dummy_sink")
    @sink_to_file(".dummy")
    def to_dummy(quantity, filename, **params) -> None:
        pass

    @register_output_quantity
    def dummy(market, **kwargs):
        return xr.DataArray()

    mca = setup_mca_test(
        tmp_path,
        dict(quantity="dummy", sink="to_dummy", filename="{path}/{Quantity}{suffix}"),
    )
    assert mca.outputs(mca.market)[0] == tmp_path / "model" / "Dummy.dummy"


def test_register_output_quantity_cache():
    from muse.outputs.cache import OUTPUT_QUANTITIES, register_cached_quantity

    @register_cached_quantity
    def dummy_quantity(*args):
        pass

    assert OUTPUT_QUANTITIES[dummy_quantity.__name__] == dummy_quantity


class TestOutputCache:
    @fixture
    def output_params(self):
        return [dict(quantity="height"), dict(quantity="width")]

    @fixture
    def output_quantities(self, output_params):
        quantities = {q["quantity"]: lambda _: None for q in output_params}
        quantities["depth"] = lambda _: None
        return quantities

    @fixture
    def topic(self):
        return "BBC Muse"

    @patch("pubsub.pub.subscribe")
    @patch("muse.outputs.sector._factory")
    def test_init(
        self, mock_factory, mock_subscribe, output_params, output_quantities, topic
    ):
        from muse.outputs.cache import OutputCache

        output_cache = OutputCache(
            *output_params, output_quantities=output_quantities, topic=topic
        )
        assert mock_factory.call_count == len(output_params)
        mock_subscribe.assert_called_once_with(output_cache.cache, topic)

    @patch("pubsub.pub.subscribe")
    @patch("muse.outputs.sector._factory")
    def test_cache(
        self, mock_factory, mock_subscribe, output_params, output_quantities, topic
    ):
        from muse.outputs.cache import OutputCache

        output_cache = OutputCache(
            *output_params, output_quantities=output_quantities, topic=topic
        )
        output_cache.cache(dict(height=xr.DataArray(), depth=xr.DataArray()))
        assert len(output_cache.to_save.get("height")) == 1
        assert len(output_cache.to_save.get("depth", [])) == 0

    @patch("pubsub.pub.subscribe")
    @patch("muse.outputs.sector._factory")
    def test_consolidate_cache(
        self, mock_factory, mock_subscribe, output_params, output_quantities, topic
    ):
        from muse.outputs.cache import OutputCache

        output_cache = OutputCache(
            *output_params, output_quantities=output_quantities, topic=topic
        )
        output_cache.cache(dict(height=xr.DataArray()))
        output_cache.consolidate_cache(2042)
        output_cache.factory["height"].assert_called_once()


@patch("pubsub.pub.sendMessage")
@patch("muse.outputs.cache.match_quantities")
def test_cache_quantity(mock_match, mock_send):
    from muse.outputs.cache import CACHE_TOPIC_CHANNEL, cache_quantity

    result = {"mass": 42}
    mock_match.return_value = result

    def verify_message_sent():
        mock_send.assert_called_once_with(CACHE_TOPIC_CHANNEL, data=result)
        mock_send.reset_mock()

    cache_quantity(**result)
    verify_message_sent()

    with raises(ValueError):
        cache_quantity(function=lambda: None, mass=42)

    with raises(ValueError):

        @cache_quantity
        def fun():
            pass

    @cache_quantity(quantity="mass")
    def fun2():
        return 42

    fun2()
    mock_match.assert_called_once_with("mass", 42)
    verify_message_sent()


def test_match_quantities():
    import xarray as xr

    from muse.outputs.cache import match_quantities

    def assert_equal(a: dict[str, xr.DataArray], b: dict[str, xr.DataArray]):
        assert set(a.keys()) == set(b.keys())
        for k in a:
            xr.testing.assert_equal(a[k], b[k])

    # Test single quantity with DataArray
    q = "mass"
    da = xr.DataArray(name=q)
    ds = xr.Dataset({q: da})
    assert_equal(match_quantities(quantity=q, data=da), {q: da})
    assert_equal(match_quantities(quantity=q, data=ds), {q: da})

    # Test multiple quantities with Dataset
    p = "height"
    ds = xr.Dataset({q: da, p: da, "rubish": da})
    assert_equal(match_quantities(quantity=[q, p], data=ds), {q: da, p: da})
    assert_equal(match_quantities(quantity=[q, p], data=[da, da]), {q: da, p: da})

    # Test error cases
    with raises(ValueError):
        match_quantities(quantity=[q, p], data=[da])
    with raises(TypeError):
        match_quantities(quantity=[q, p], data=42)


@patch("muse.outputs.cache.extract_agents_internal")
def test_extract_agents(mock_extract):
    from muse.outputs.cache import extract_agents

    sectors = set((1, 2, 3))
    mock_extract.side_effect = [{s: None} for s in sectors]

    actual = extract_agents(sectors)

    assert set(actual.keys()) == sectors
    assert set(actual.values()) == set((None,))


def test_extract_agents_internal(newcapa_agent, retro_agent):
    """Test internal agent extraction."""
    from types import SimpleNamespace

    from muse.outputs.cache import extract_agents_internal

    def setup_agent(agent, name):
        agent.name = name
        return agent

    agents = [setup_agent(newcapa_agent, "A1"), setup_agent(retro_agent, "A2")]
    sector = SimpleNamespace(name="IT", agents=agents)

    actual = extract_agents_internal(sector)
    expected_keys = ("agent", "category", "sector", "dst_region")

    for agent in agents:
        assert agent.uuid in actual
        assert tuple(actual[agent.uuid].keys()) == expected_keys
        agent_data = actual[agent.uuid]
        assert agent_data["agent"] == agent.name
        assert agent_data["category"] == agent.category
        assert agent_data["sector"] == "IT"
        assert agent_data["dst_region"] == agent.region


def test_aggregate_cache():
    import numpy as np
    from pandas.testing import assert_frame_equal

    from muse.outputs.cache import _aggregate_cache

    quantity = "height"
    a = xr.DataArray(np.ones((3, 4, 5)), name=quantity)
    b = a.copy()
    b[0, 0, 0] = 0

    def to_df(arr):
        return arr.to_dataframe().reset_index().astype(float)

    actual = _aggregate_cache(quantity, [a, b])
    assert_frame_equal(actual, to_df(b))

    actual = _aggregate_cache(quantity, [b, a])
    assert_frame_equal(actual, to_df(a))

    c = a.copy()
    c.assign_coords(dim_0=c.dim_0.data * 10)
    dc, da = map(to_df, [c, a])

    actual = _aggregate_cache(quantity, [c, a])
    expected = pd.DataFrame.merge(dc, da, how="outer").astype(float)
    assert_frame_equal(actual, expected)


def test_consolidate_quantity(newcapa_agent, retro_agent):
    """Test consolidation of quantity data with agent information."""
    from types import SimpleNamespace

    from muse.outputs.cache import consolidate_quantity, extract_agents_internal

    def setup_agent(agent, name, category):
        agent.name = name
        agent.category = category
        return agent

    newcapa_agent = setup_agent(newcapa_agent, "A1", "newcapa")
    retro_agent = setup_agent(retro_agent, "A2", "retro")
    sector = SimpleNamespace(name="IT", agents=[newcapa_agent, retro_agent])
    agents = extract_agents_internal(sector)

    def create_agent_array(agent_uuid, modify_first=False):
        arr = xr.DataArray(
            np.ones((3, 4, 5)),
            dims=("agent", "replacement", "asset"),
            coords={"agent": [agent_uuid] * 3},
            name="height",
        )
        if modify_first:
            arr[0, 0, 0] = 0
        return arr

    a = create_agent_array(newcapa_agent.uuid)
    b = create_agent_array(retro_agent.uuid, modify_first=True)

    actual = consolidate_quantity("height", [a, b], agents)
    assert set(actual.columns) == {
        "agent",
        "asset",
        "category",
        "sector",
        "dst_region",
        "height",
        "replacement",
        "sector",
    }
    assert all(
        name in (newcapa_agent.name, retro_agent.name) for name in actual.agent.unique()
    )



"""Valuation functions for replacement technologies.

.. currentmodule:: muse.objectives

Objectives are used to compare replacement technologies. They should correspond to
a single well defined economic concept. Multiple objectives can later be combined
via decision functions.

Objectives should be registered via the
:py:func:`@register_objective<register_objective>` decorator. This makes it possible to
refer to them by name in agent input files, and nominally to set extra input parameters.

The :py:func:`factory` function creates a function that calls all objectives defined in
its input argument and returns a dataset with each objective as a separate data array.

Objectives are not expected to modify their arguments. Furthermore they should
conform the following signatures:

.. code-block:: Python

    @register_objective
    def comfort(
        technologies: xr.Dataset,
        demand: xr.DataArray,
        prices: xr.DataArray,
        **kwargs
    ) -> xr.DataArray:
        pass

Arguments:
    technologies: A data set characterising the technologies from which the
        agent can draw assets. This has been pre-filtered according to the agent's
        search space.
    demand: Demand to fulfill.
    prices: Commodity prices.
    kwargs: Extra input parameters. These parameters are expected to be set from the
        input file.

        .. warning::

            The standard :ref:`agent csv file<inputs-agents>` does not allow to set
            these parameters.

Returns:
    A DataArray with at least two dimension corresponding to `replacement` and `asset`.
    A `timeslice` dimension may also be present.
"""

from __future__ import annotations

__all__ = [
    "capacity_to_service_demand",
    "capital_costs",
    "comfort",
    "efficiency",
    "emission_cost",
    "equivalent_annual_cost",
    "factory",
    "fixed_costs",
    "fuel_consumption_cost",
    "lifetime_levelized_cost_of_energy",
    "net_present_value",
    "register_objective",
]

from collections.abc import Mapping, MutableMapping, Sequence
from typing import Any, Callable, Union

import numpy as np
import xarray as xr
from mypy_extensions import KwArg

from muse.outputs.cache import cache_quantity
from muse.registration import registrator
from muse.timeslices import broadcast_timeslice, distribute_timeslice, drop_timeslice
from muse.utilities import check_dimensions

OBJECTIVE_SIGNATURE = Callable[
    [xr.Dataset, xr.DataArray, xr.DataArray, KwArg(Any)], xr.DataArray
]
"""Objectives signature."""

OBJECTIVES: MutableMapping[str, OBJECTIVE_SIGNATURE] = {}
"""Dictionary of objectives when selecting replacement technology."""


def objective_factory(settings=Union[str, Mapping]):
    from functools import partial

    if isinstance(settings, str):
        params = dict(name=settings)
    else:
        params = dict(**settings)
    name = params.pop("name")
    function = OBJECTIVES[name]
    return partial(function, **params)


def factory(
    settings: str | Mapping | Sequence[str | Mapping] = "LCOE",
) -> Callable:
    """Creates a function computing multiple objectives.

    The input can be a single objective defined by its name alone. Or it can be a single
    objective defined by a dictionary which must include at least a "name" item, as well
    as any extra parameters to pass to the objective. Or it can be a sequence of
    objectives defined by name or by dictionary.
    """
    from logging import getLogger

    if isinstance(settings, str):
        params: list[dict] = [{"name": settings}]
    elif isinstance(settings, Mapping):
        params = [dict(**settings)]
    else:
        params = [
            {"name": param} if isinstance(param, str) else dict(**param)
            for param in settings
        ]

    if len(set(param["name"] for param in params)) != len(params):
        msg = (
            "The same objective is named twice."
            " The result may be undefined if parameters differ."
        )
        getLogger(__name__).critical(msg)

    functions = [(param["name"], objective_factory(param)) for param in params]

    def objectives(
        technologies: xr.Dataset,
        demand: xr.DataArray,
        prices: xr.DataArray,
        timeslice_level: str | None = None,
        *args,
        **kwargs,
    ) -> xr.Dataset:
        result = xr.Dataset()
        for name, objective in functions:
            obj = objective(
                technologies=technologies,
                demand=demand,
                prices=prices,
                timeslice_level=timeslice_level,
                *args,
                **kwargs,
            )
            if "timeslice" not in obj.dims:
                obj = broadcast_timeslice(obj, level=timeslice_level)
            if "timeslice" in result.dims:
                obj = drop_timeslice(obj)
            result[name] = obj
        return result

    return objectives


@registrator(registry=OBJECTIVES, loglevel="info")
def register_objective(function: OBJECTIVE_SIGNATURE):
    """Decorator to register a function as a objective.

    Registers a function as a objective so that it can be applied easily
    when sorting technologies one against the other.

    The input name is expected to be in lower_snake_case, since it ought to be a
    python function. CamelCase, lowerCamelCase, and kebab-case names are also
    registered.
    """
    from functools import wraps

    @wraps(function)
    def decorated_objective(
        technologies: xr.Dataset, demand: xr.DataArray, *args, **kwargs
    ) -> xr.DataArray:
        from logging import getLogger

        # Check inputs
        check_dimensions(
            demand, ["asset", "timeslice", "commodity"], optional=["region"]
        )
        check_dimensions(
            technologies, ["replacement", "commodity"], optional=["timeslice", "asset"]
        )

        # Calculate objective
        result = function(technologies, demand, *args, **kwargs)
        result.name = function.__name__

        # Check result
        dtype = result.values.dtype
        if not (np.issubdtype(dtype, np.number) or np.issubdtype(dtype, np.bool_)):
            msg = f"dtype of objective {function.__name__} is not a number ({dtype})"
            getLogger(function.__module__).warning(msg)
        check_dimensions(result, ["replacement", "asset"], optional=["timeslice"])

        cache_quantity(**{result.name: result})
        return result

    return decorated_objective


@register_objective
def comfort(
    technologies: xr.Dataset,
    demand: xr.DataArray,
    *args,
    **kwargs,
) -> xr.DataArray:
    """Comfort value provided by technologies."""
    result = xr.broadcast(technologies.comfort, demand.asset)[0]
    return result


@register_objective
def efficiency(
    technologies: xr.Dataset,
    demand: xr.DataArray,
    *args,
    **kwargs,
) -> xr.DataArray:
    """Efficiency of the technologies."""
    result = xr.broadcast(technologies.efficiency, demand.asset)[0]
    return result


@register_objective(name="capacity")
def capacity_to_service_demand(
    technologies: xr.Dataset,
    demand: xr.DataArray,
    timeslice_level: str | None = None,
    *args,
    **kwargs,
) -> xr.DataArray:
    """Minimum capacity required to fulfill the demand."""
    from muse.quantities import capacity_to_service_demand

    return capacity_to_service_demand(
        demand=demand, technologies=technologies, timeslice_level=timeslice_level
    )


@register_objective
def capacity_in_use(
    technologies: xr.Dataset,
    demand: xr.DataArray,
    *args,
    **kwargs,
):
    from muse.commodities import is_enduse
    from muse.timeslices import TIMESLICE

    enduses = is_enduse(technologies.comm_usage.sel(commodity=demand.commodity))
    return (
        (demand.sel(commodity=enduses).sum("commodity") / TIMESLICE).sum("timeslice")
        * TIMESLICE.sum()
        / technologies.utilization_factor
    )


@register_objective
def consumption(
    technologies: xr.Dataset,
    demand: xr.DataArray,
    prices: xr.DataArray,
    timeslice_level: str | None = None,
    *args,
    **kwargs,
) -> xr.DataArray:
    """Commodity consumption when fulfilling the whole demand.

    Currently, the consumption is implemented for commodity_max == +infinity.
    """
    from muse.quantities import consumption
    from muse.timeslices import broadcast_timeslice, distribute_timeslice

    capacity = capacity_to_service_demand(
        technologies, demand, timeslice_level=timeslice_level
    )
    production = (
        broadcast_timeslice(capacity, level=timeslice_level)
        * distribute_timeslice(technologies.fixed_outputs, level=timeslice_level)
        * broadcast_timeslice(technologies.utilization_factor, level=timeslice_level)
    )
    consump = consumption(
        technologies=technologies,
        prices=prices,
        production=production,
        timeslice_level=timeslice_level,
    )
    return consump.sum("commodity")


@register_objective
def fixed_costs(
    technologies: xr.Dataset,
    demand: xr.DataArray,
    timeslice_level: str | None = None,
    *args,
    **kwargs,
) -> xr.DataArray:
    r"""Fixed costs associated with a technology.

    Given a factor :math:`\alpha` and an  exponent :math:`\beta`, the fixed costs
    :math:`F` are computed from the :py:func:`capacity fulfilling the current demand
    <capacity_to_service_demand>` :math:`C` as:

    .. math::

        F = \alpha * C^\beta

    :math:`\alpha` and :math:`\beta` are "fix_par" and "fix_exp" in
    :ref:`inputs-technodata`, respectively.
    """
    from muse.costs import fixed_costs

    capacity = capacity_to_service_demand(
        technologies, demand, timeslice_level=timeslice_level
    )
    result = fixed_costs(technologies, capacity)
    return result


@register_objective
def capital_costs(
    technologies: xr.Dataset,
    demand: xr.Dataset,
    timeslice_level: str | None = None,
    *args,
    **kwargs,
) -> xr.DataArray:
    """Capital costs for input technologies."""
    from muse.costs import capital_costs

    capacity = capacity_to_service_demand(
        technologies, demand, timeslice_level=timeslice_level
    )
    result = capital_costs(technologies, capacity, method="lifetime")
    result = xr.broadcast(result, demand.asset)[0]
    return result


@register_objective(name="emissions")
def emission_cost(
    technologies: xr.Dataset,
    demand: xr.DataArray,
    prices: xr.DataArray,
    timeslice_level: str | None = None,
    *args,
    **kwargs,
) -> xr.DataArray:
    r"""Emission cost for each technology when fulfilling whole demand.

    Given the demand share :math:`D`, the emissions per amount produced :math:`E`, and
    the prices per emittant :math:`P`, then emissions costs :math:`C` are computed
    as:

    .. math::

        C = \sum_s \left(\sum_cD\right)\left(\sum_cEP\right),

    with :math:`s` the timeslices and :math:`c` the commodity.
    """
    from muse.costs import environmental_costs

    capacity = capacity_to_service_demand(
        technologies, demand, timeslice_level=timeslice_level
    )
    production = (
        broadcast_timeslice(capacity, level=timeslice_level)
        * distribute_timeslice(technologies.fixed_outputs, level=timeslice_level)
        * broadcast_timeslice(technologies.utilization_factor, level=timeslice_level)
    )
    result = environmental_costs(technologies, prices, production)
    return result


@register_objective
def fuel_consumption_cost(
    technologies: xr.Dataset,
    demand: xr.DataArray,
    prices: xr.DataArray,
    timeslice_level: str | None = None,
    *args,
    **kwargs,
):
    """Cost of fuels when fulfilling whole demand."""
    from muse.costs import fuel_costs
    from muse.quantities import consumption
    from muse.timeslices import broadcast_timeslice, distribute_timeslice

    capacity = capacity_to_service_demand(
        technologies, demand, timeslice_level=timeslice_level
    )
    production = (
        broadcast_timeslice(capacity, level=timeslice_level)
        * distribute_timeslice(technologies.fixed_outputs, level=timeslice_level)
        * broadcast_timeslice(technologies.utilization_factor, level=timeslice_level)
    )
    consump = consumption(
        technologies=technologies,
        prices=prices,
        production=production,
        timeslice_level=timeslice_level,
    )
    result = fuel_costs(technologies, prices, consump)
    return result


@register_objective(name=["ALCOE"])
def annual_levelized_cost_of_energy(
    technologies: xr.Dataset,
    demand: xr.DataArray,
    prices: xr.DataArray,
    timeslice_level: str | None = None,
    *args,
    **kwargs,
):
    """Annual cost of energy (LCOE) of technologies - not dependent on production.

    It needs to be used for trade agents where the actual service is unknown. It follows
    the `simplified LCOE` given by NREL.

    See :py:func:`muse.costs.levelized_cost_of_energy` for more details.

    """
    from muse.costs import levelized_cost_of_energy as LCOE
    from muse.quantities import consumption

    capacity = capacity_to_service_demand(
        technologies, demand, timeslice_level=timeslice_level
    )
    production = (
        broadcast_timeslice(capacity, level=timeslice_level)
        * distribute_timeslice(technologies.fixed_outputs, level=timeslice_level)
        * broadcast_timeslice(technologies.utilization_factor, level=timeslice_level)
    )
    consump = consumption(
        technologies=technologies,
        prices=prices,
        production=production,
        timeslice_level=timeslice_level,
    )

    results = LCOE(
        technologies=technologies,
        prices=prices,
        capacity=capacity,
        production=production,
        consumption=consump,
        method="annual",
        aggregate_timeslices=True,
    )
    return results


@register_objective(name=["LCOE", "LLCOE"])
def lifetime_levelized_cost_of_energy(
    technologies: xr.Dataset,
    demand: xr.DataArray,
    prices: xr.DataArray,
    timeslice_level: str | None = None,
    *args,
    **kwargs,
):
    """Levelized cost of energy (LCOE) of technologies over their lifetime.

    See :py:func:`muse.costs.levelized_cost_of_energy` for more details.

    The LCOE is set to zero for those timeslices where the production is zero, normally
    due to a zero utilization factor.
    """
    from muse.costs import levelized_cost_of_energy as LCOE
    from muse.quantities import capacity_to_service_demand, consumption

    capacity = capacity_to_service_demand(
        technologies=technologies, demand=demand, timeslice_level=timeslice_level
    )
    production = (
        broadcast_timeslice(capacity, level=timeslice_level)
        * distribute_timeslice(technologies.fixed_outputs, level=timeslice_level)
        * broadcast_timeslice(technologies.utilization_factor, level=timeslice_level)
    )
    consump = consumption(
        technologies=technologies,
        prices=prices,
        production=production,
        timeslice_level=timeslice_level,
    )

    results = LCOE(
        technologies=technologies,
        prices=prices,
        capacity=capacity,
        production=production,
        consumption=consump,
        method="lifetime",
        aggregate_timeslices=True,
    )
    return results


@register_objective(name="NPV")
def net_present_value(
    technologies: xr.Dataset,
    demand: xr.DataArray,
    prices: xr.DataArray,
    timeslice_level: str | None = None,
    *args,
    **kwargs,
):
    """Net present value (NPV) of the relevant technologies.

    See :py:func:`muse.costs.net_present_value` for more details.
    """
    from muse.costs import net_present_value as NPV
    from muse.quantities import capacity_to_service_demand, consumption

    capacity = capacity_to_service_demand(
        technologies=technologies, demand=demand, timeslice_level=timeslice_level
    )
    production = (
        broadcast_timeslice(capacity, level=timeslice_level)
        * distribute_timeslice(technologies.fixed_outputs, level=timeslice_level)
        * broadcast_timeslice(technologies.utilization_factor, level=timeslice_level)
    )
    consump = consumption(
        technologies=technologies,
        prices=prices,
        production=production,
        timeslice_level=timeslice_level,
    )

    results = NPV(
        technologies=technologies,
        prices=prices,
        capacity=capacity,
        production=production,
        consumption=consump,
        aggregate_timeslices=True,
    )
    return results


@register_objective(name="NPC")
def net_present_cost(
    technologies: xr.Dataset,
    demand: xr.DataArray,
    prices: xr.DataArray,
    timeslice_level: str | None = None,
    *args,
    **kwargs,
):
    """Net present cost (NPC) of the relevant technologies.

    See :py:func:`muse.costs.net_present_cost` for more details.
    """
    from muse.costs import net_present_cost as NPC
    from muse.quantities import capacity_to_service_demand, consumption

    capacity = capacity_to_service_demand(
        technologies=technologies, demand=demand, timeslice_level=timeslice_level
    )
    production = (
        broadcast_timeslice(capacity, level=timeslice_level)
        * distribute_timeslice(technologies.fixed_outputs, level=timeslice_level)
        * broadcast_timeslice(technologies.utilization_factor, level=timeslice_level)
    )
    consump = consumption(
        technologies=technologies,
        prices=prices,
        production=production,
        timeslice_level=timeslice_level,
    )

    results = NPC(
        technologies=technologies,
        prices=prices,
        capacity=capacity,
        production=production,
        consumption=consump,
        aggregate_timeslices=True,
    )
    return results


@register_objective(name="EAC")
def equivalent_annual_cost(
    technologies: xr.Dataset,
    demand: xr.DataArray,
    prices: xr.DataArray,
    timeslice_level: str | None = None,
    *args,
    **kwargs,
):
    """Equivalent annual costs (or annualized cost) of a technology.

    See :py:func:`muse.costs.equivalent_annual_cost` for more details.
    """
    from muse.costs import equivalent_annual_cost as EAC
    from muse.quantities import capacity_to_service_demand, consumption

    capacity = capacity_to_service_demand(
        technologies=technologies, demand=demand, timeslice_level=timeslice_level
    )
    production = (
        broadcast_timeslice(capacity, level=timeslice_level)
        * distribute_timeslice(technologies.fixed_outputs, level=timeslice_level)
        * broadcast_timeslice(technologies.utilization_factor, level=timeslice_level)
    )
    consump = consumption(
        technologies=technologies,
        prices=prices,
        production=production,
        timeslice_level=timeslice_level,
    )

    results = EAC(
        technologies=technologies,
        prices=prices,
        capacity=capacity,
        production=production,
        consumption=consump,
        aggregate_timeslices=True,
    )
    return results



r"""Investment constraints.

Constraints on investments ensure that investments match some given criteria. For
instance, the constraints could ensure that only so much of a new asset can be built
every year.

Functions to compute constraints should be registered via the decorator
:py:meth:`~muse.constraints.register_constraints`. This registration step makes it
possible for constraints to be declared in the TOML file.

Generally, LP solvers accept linear constraints defined as:

.. math::

    A x \\leq b

with :math:`A` a matrix, :math:`x` the decision variables, and :math:`b` a vector.
However, these quantities are dimensionless. They do no have timeslices, assets, or
replacement technologies, or any other dimensions that users have set up in their model.
The crux is to translate from MUSE's data-structures to a consistent dimensionless
format.

In MUSE, users can register constraints functions that return fully dimensional
quantities. The matrix operator is split over the capacity decision variables and the
production decision variables:

.. math::

    A_c .* x_c + A_p .* x_p \\leq b

The operator :math:`.*` means the standard elementwise multiplication of xarray,
including automatic broadcasting (adding missing dimensions by repeating the smaller
matrix along the missing dimension).  Constraint functions return the three quantities
:math:`A_c`, :math:`A_p`, and :math:`b`. These three quantities will often not have the
same dimension. E.g. one might include timeslices where another might not. The
transformation from :math:`A_c`, :math:`A_p`, :math:`b` to :math:`A` and :math:`b`
happens as described below.

- :math:`b` remains the same. It defines the rows of :math:`A`.
- :math:`x_c` and :math:`x_p` are concatenated one on top of the other and define the
  columns of :math:`A`.
- :math:`A` is split into a left submatrix for capacities and a right submatrix for
  production, following the concatenation of :math:`x_c` and :math:`x_p`
- Any dimension in :math:`A_c .* x_c` (:math:`A_p .* x_p`) that is also in :math:`b`
  defines diagonal entries into the left (right) submatrix of :math:`A`.
- Any dimension in :math:`A_c .* x_c` (:math:`A_p .* x_b`) and missing from
  :math:`b` is reduced by summation over a row in the left (right) submatrix of
  :math:`A`. In other words, those dimensions become part of a standard tensor
  reduction or matrix multiplication.

There are two additional rules. However, they are likely to be the result of an
inefficient definition of :math:`A_c`, :math:`A_p` and :math:`b`.

- Any dimension in :math:`A_c` (:math:`A_b`) that is neither in :math:`b` nor in
  :math:`x_c` (:math:`x_p`) is reduced by summation before consideration for the
  elementwise multiplication. For instance, if :math:`d` is such a dimension, present
  only in :math:`A_c`, then the problem becomes :math:`(\\sum_d A_c) .* x_c + A_p .* x_p
  \\leq b`.
- Any dimension missing from :math:`A_c .* x_c` (:math:`A_p .* x_p`) and present in
  :math:`b` is added by repeating the resulting row in :math:`A`.

Constraints are registered using the decorator
:py:meth:`~muse.constraints.register_constraints`. The decorated functions must follow
the following signature:

.. code-block:: python

    @register_constraints
    def constraints(
        demand: xr.DataArray,
        capacity: xr.DataArray,
        search_space: xr.DataArray,
        technologies: xr.Dataset,
        **kwargs,
    ) -> Constraint:
        pass

demand:
    The demand for the sectors products in the investment year. In practice it is a
    demand share obtained in :py:mod:`~muse.demand_share`. It is a data-array with
    dimensions including `asset`, `commodity`, `timeslice`.
capacity:
    A data-array with dimensions `technology` and `year` defining the existing capacity
    of each technology in the current year and investment year.
search_space:
    A matrix `asset` vs `replacement` technology defining which replacement technologies
    will be considered for each existing asset.
technologies:
    Technodata characterizing the competing technologies in the investment year.
``**kwargs``:
    Any other parameter.
"""

from __future__ import annotations

from collections.abc import Mapping, MutableMapping, Sequence
from enum import Enum, auto
from typing import (
    Any,
    Callable,
    Optional,
    Union,
    cast,
)

import numpy as np
import pandas as pd
import xarray as xr
from mypy_extensions import KwArg

from muse.registration import registrator
from muse.timeslices import broadcast_timeslice, distribute_timeslice

CAPACITY_DIMS = "asset", "replacement", "region"
"""Default dimensions for capacity decision variables."""
PRODUCT_DIMS = "commodity", "timeslice", "region"
"""Default dimensions for product decision variables."""


class ConstraintKind(Enum):
    EQUALITY = auto()
    UPPER_BOUND = auto()
    LOWER_BOUND = auto()


Constraint = xr.Dataset
"""An investment constraint :math:`A * x ~ b`

Where :math:`~` is one of :math:`=,\\leq,\\geq`.

A constraint should contain a data-array `b` corresponding to right-hand-side vector
of the constraint. It should also contain a data-array `capacity` corresponding to the
left-hand-side matrix operator which will be applied to the capacity-related decision
variables.  It should contain a similar matrix `production` corresponding to
the left-hand-side matrix operator which will be applied to the production-related
decision variables. Should any of these three objects be missing, they default to the
scalar 0. Finally, the constraint should contain an attribute `kind` of type
:py:class:`ConstraintKind` defining the operation. If it is missing, it defaults to an
upper bound constraint.
"""


CONSTRAINT_SIGNATURE = Callable[
    [xr.DataArray, xr.DataArray, xr.DataArray, xr.Dataset, KwArg(Any)],
    Optional[Constraint],
]
"""Basic signature for functions producing constraints.

.. note::

    A constraint can return `None`, in which case it is ignored. This makes it simple to
    add constraints that are only used if some condition is met, e.g. minimum service
    conditions are defined in the technodata.
"""
CONSTRAINTS: MutableMapping[str, CONSTRAINT_SIGNATURE] = {}
"""Registry of constraint functions."""


@registrator(registry=CONSTRAINTS)
def register_constraints(function: CONSTRAINT_SIGNATURE) -> CONSTRAINT_SIGNATURE:
    """Registers a constraint with MUSE.

    See :py:mod:`muse.constraints`.
    """
    from functools import wraps

    @wraps(function)
    def decorated(
        demand: xr.DataArray,
        capacity: xr.DataArray,
        search_space: xr.DataArray,
        technologies: xr.Dataset,
        **kwargs,
    ) -> Constraint | None:
        """Computes and standardizes a constraint."""
        # Check inputs
        assert "year" not in technologies.dims
        assert "year" not in demand.dims
        assert len(capacity.year) == 2  # current year and investment year

        # Calculate constraint
        constraint = function(  # type: ignore
            demand,
            capacity=capacity,
            search_space=search_space,
            technologies=technologies,
            **kwargs,
        )

        if constraint is not None:
            # Check constraint
            if "b" not in constraint.data_vars:
                raise RuntimeError("Constraint must contain a right-hand-side vector")
            if (
                "capacity" not in constraint.data_vars
                and "production" not in constraint.data_vars
            ):
                raise RuntimeError("Constraint must contain a left-hand-side matrix")
            if "kind" not in constraint.attrs:
                raise RuntimeError("Constraint must contain a kind attribute")

            # Standardize constraint
            if "capacity" not in constraint.data_vars:
                constraint["capacity"] = 0
            if "production" not in constraint.data_vars:
                constraint["production"] = 0
            if "name" not in constraint.data_vars and "name" not in constraint.attrs:
                constraint.attrs["name"] = function.__name__

            # ensure that the constraint and the search space match
            dims = [d for d in constraint.dims if d in search_space.dims]
            constraint = constraint.sel({k: search_space[k] for k in dims})

        return constraint

    return decorated


def factory(
    settings: str | Mapping | Sequence[str] | Sequence[str | Mapping] | None = None,
) -> Callable:
    """Creates a list of constraints from standard settings.

    The standard settings can be a string naming the constraint, a dictionary including
    at least "name", or a list of strings and dictionaries.
    """
    from functools import partial

    if not settings:
        settings = (
            "max_production",
            "max_capacity_expansion",
            "demand",
            "search_space",
            "minimum_service",
            "demand_limiting_capacity",
        )

    def normalize(x) -> MutableMapping:
        return dict(name=x) if isinstance(x, str) else x

    if isinstance(settings, (str, Mapping)):
        settings = cast(Union[Sequence[str], Sequence[Mapping]], [settings])
    parameters = [normalize(x) for x in settings]
    names = [x.pop("name") for x in parameters]

    constraint_closures = [
        partial(CONSTRAINTS[name], **param) for name, param in zip(names, parameters)
    ]

    def constraints(
        demand: xr.DataArray,
        capacity: xr.DataArray,
        search_space: xr.DataArray,
        technologies: xr.Dataset,
        **kwargs,
    ) -> list[Constraint]:
        constraints = [
            function(
                demand,
                capacity=capacity,
                search_space=search_space,
                technologies=technologies,
                **kwargs,
            )
            for function in constraint_closures
        ]
        return [constraint for constraint in constraints if constraint is not None]

    return constraints


@register_constraints
def max_capacity_expansion(
    demand: xr.DataArray,
    capacity: xr.DataArray,
    search_space: xr.DataArray,
    technologies: xr.Dataset,
    **kwargs,
) -> Constraint | None:
    r"""Max-capacity addition, max-capacity growth, and capacity limits constraints.

    Limits by how much the capacity of each technology owned by an agent can grow in
    a given year. This is a constraint on the agent's ability to invest in a
    technology.

    Let :math:`L_t^r(y)` be the total capacity limit for a given year, technology,
    and region. :math:`G_t^r(y)` is the maximum growth. And :math:`W_t^r(y)` is
    the maximum additional capacity. :math:`y=y_0` is the current year and
    :math:`y=y_1` is the year marking the end of the investment period.

    Let :math:`\mathcal{A}^{i, r}_{t, \iota}(y)` be the current assets, before
    investment, and let :math:`\Delta\mathcal{A}^{i,r}_t` be the future investments.
    The the constraint on agent :math:`i` are given as:

    .. math::

        L_t^r(y_0) - \sum_\iota \mathcal{A}^{i, r}_{t, \iota}(y_1)
            \geq \Delta\mathcal{A}^{i,r}_t

        (y_1 - y_0 + 1) G_t^r(y_0) \sum_\iota \mathcal{A}^{i, r}_{t, \iota}(y_0)
            - \sum_\iota \mathcal{A}^{i, r}_{t, \iota}(y_1)
            \geq \Delta\mathcal{A}^{i,r}_t

        (y_1 - y_0)W_t^r(y_0) \geq  \Delta\mathcal{A}^{i,r}_t

    The three constraints are combined into a single one which is returned as the
    maximum capacity expansion, :math:`\Gamma_t^{r, i}`. The maximum capacity
    expansion cannot impose negative investments:
    Maximum capacity addition:

        .. math::

            \Gamma_t^{r, i} \geq 0

    :math:`L_t^r(y)`, :math:`G_t^r(y)` and :math:`W_t^r(y)` default to np.inf if
    not provided (i.e. no constraint).
    If all three parameters are not provided, no constraint is applied (returns None).
    """
    # If all three parameters are missing, don't apply the constraint
    if not any(
        param in technologies
        for param in (
            "max_capacity_addition",
            "total_capacity_limit",
            "max_capacity_growth",
        )
    ):
        return None

    # case with technology and region in asset dimension
    if capacity.region.dims != ():
        names = [u for u in capacity.asset.coords if capacity[u].dims == ("asset",)]
        index = pd.MultiIndex.from_arrays(
            [capacity[u].values for u in names], names=names
        )
        mindex_coords = xr.Coordinates.from_pandas_multiindex(index, "asset")
        capacity = capacity.drop_vars(names).assign_coords(mindex_coords)
        capacity = capacity.unstack("asset", fill_value=0).rename(
            technology=search_space.replacement.name
        )
    # case with only technology in asset dimension
    else:
        capacity = cast(xr.DataArray, capacity.set_index(asset="technology")).rename(
            asset=search_space.replacement.name
        )
    capacity = capacity.reindex_like(search_space.replacement, fill_value=0)

    replacement = search_space.replacement
    replacement = replacement.drop_vars(
        [u for u in replacement.coords if u not in replacement.dims]
    )
    techs = technologies.sel(technology=replacement).drop_vars("technology")
    regions = getattr(capacity, "region", None)
    if regions is not None and "region" in technologies.dims:
        techs = techs.sel(region=regions)

    # Existing and forecasted capacity
    initial = capacity.isel(year=0, drop=True)
    forecasted = capacity.isel(year=1, drop=True)
    time_frame = int(capacity.year[1] - capacity.year[0])

    # MaxCapacityAddition constraint
    if "max_capacity_addition" in techs:
        add_cap = techs.max_capacity_addition * time_frame
    else:
        add_cap = np.inf

    # TotalCapacityLimit constraint
    if "total_capacity_limit" in techs:
        limit = techs.total_capacity_limit
        total_cap = (limit - forecasted).clip(min=0)
    else:
        total_cap = np.inf

    # MaxCapacityGrowth constraint
    if "max_capacity_growth" in techs:
        seed = techs.get("growth_seed", 1)
        seeded_initial = np.maximum(initial, seed)
        growth_cap = (
            (seeded_initial * (techs.max_capacity_growth + 1) ** time_frame)
            - forecasted
        ).clip(min=0)
    else:
        growth_cap = np.inf

    # Take the most restrictive constraint
    b = np.minimum(np.minimum(add_cap, total_cap), growth_cap)

    # np.inf values are not allowed in the final constraint - raise error
    # Will happen if user provides "inf" for all three parameters for any technology
    if np.isinf(b).any():
        inf_replacements = b.replacement[np.isinf(b)].values
        raise ValueError(
            "Capacity growth constraint cannot be infinite. "
            f"Check growth constraint parameters for technologies: {inf_replacements}"
        )

    # np.nan values are also not allowed. This shouldn't happen but just in case
    if np.isnan(b).any():
        nan_replacements = b.replacement[np.isnan(b)].values
        raise ValueError(
            "Capacity growth constraint cannot be NaN. "
            f"Check growth constraint parameters for technologies: {nan_replacements}"
        )

    if b.region.dims == ():
        capa = 1
    elif "dst_region" in b.dims:
        b = b.rename(region="src_region")
        capa = search_space.agent.region == b.src_region

    return xr.Dataset(
        dict(b=b, capacity=capa),
        attrs=dict(kind=ConstraintKind.UPPER_BOUND, name="max capacity expansion"),
    )


@register_constraints
def demand(
    demand: xr.DataArray,
    capacity: xr.DataArray,
    search_space: xr.DataArray,
    technologies: xr.Dataset,
    **kwargs,
) -> Constraint:
    """Constraints production to meet demand."""
    if "region" in demand.dims and "dst_region" in technologies.dims:
        demand = demand.rename(region="dst_region")
    return xr.Dataset(
        dict(b=demand, production=1), attrs=dict(kind=ConstraintKind.LOWER_BOUND)
    )


@register_constraints
def search_space(
    demand: xr.DataArray,
    capacity: xr.DataArray,
    search_space: xr.DataArray,
    technologies: xr.Dataset,
    **kwargs,
) -> Constraint | None:
    """Removes disabled technologies."""
    if search_space.all():
        return None
    capacity = cast(xr.DataArray, 1 - 2 * cast(np.ndarray, search_space))
    b = xr.zeros_like(capacity)
    return xr.Dataset(
        dict(b=b, capacity=capacity), attrs=dict(kind=ConstraintKind.UPPER_BOUND)
    )


@register_constraints
def max_production(
    demand: xr.DataArray,
    capacity: xr.DataArray,
    search_space: xr.DataArray,
    technologies: xr.Dataset,
    *,
    timeslice_level: str | None = None,
    **kwargs,
) -> Constraint:
    """Constructs constraint between capacity and maximum production.

    Constrains the production decision variable by the maximum production for a given
    capacity.
    """
    from xarray import ones_like, zeros_like

    kwargs = dict(commodity=demand.commodity)
    if "region" in search_space.coords and "region" in technologies.dims:
        kwargs["region"] = search_space.region
    techs = (
        technologies[["fixed_outputs", "utilization_factor"]]
        .sel(**kwargs)
        .rename(technology="replacement")
    )
    capa = distribute_timeslice(
        techs.fixed_outputs, level=timeslice_level
    ) * broadcast_timeslice(techs.utilization_factor, level=timeslice_level)
    if "asset" not in capa.dims and "asset" in search_space.dims:
        capa = capa.expand_dims(asset=search_space.asset)
    production = ones_like(capa)
    b = zeros_like(production)

    # Include maxaddition constraint in max production to match region-dst_region
    if "dst_region" in technologies.dims:
        b = b.expand_dims(dst_region=technologies.dst_region)
        capa = capa.rename(region="src_region")
        production = production.rename(region="src_region")
        maxadd = technologies.max_capacity_addition.rename(region="src_region")
        maxadd = maxadd.rename(technology="replacement")
        maxadd = maxadd.where(maxadd == 0, 0.0)
        maxadd = maxadd.where(maxadd > 0, -1.0)
        capa = capa * broadcast_timeslice(maxadd, level=timeslice_level)
        production = production * broadcast_timeslice(maxadd, level=timeslice_level)
        b = b.rename(region="src_region")

    return xr.Dataset(
        dict(capacity=-capa, production=production, b=b),
        attrs=dict(kind=ConstraintKind.UPPER_BOUND),
    )


@register_constraints
def demand_limiting_capacity(
    demand_: xr.DataArray,
    capacity: xr.DataArray,
    search_space: xr.DataArray,
    technologies: xr.Dataset,
    *,
    timeslice_level: str | None = None,
    **kwargs,
) -> Constraint:
    """Limits the maximum combined capacity to match the demand.

    This is a somewhat more restrictive constraint than the max_production constraint or
    the maximum capacity expansion. In this case, the combined new capacity of all
    assets must be sufficient to meet the demand of the most demanding timeslice, and
    no more.

    Rather than coding from scratch the constraint, we can use the max_production
    constraint and the demand constraint to construct this constraint. Starting from
    the maximum production instead of the maximum capacity ensures that the constraint
    accounts for the utilization factor of the technologies.
    """
    # We start with the maximum production constraint and the demand constraint
    capacity_constraint = max_production(
        demand_,
        capacity,
        search_space,
        technologies,
        timeslice_level=timeslice_level,
    )
    demand_constraint = demand(demand_, capacity, search_space, technologies)

    # We are interested in the demand of the demand constraint and the capacity of the
    # capacity constraint.
    b = demand_constraint.b
    capacity = -capacity_constraint.capacity

    # Drop 'year' so there's no conflict with the 'year' in the capacity constraint
    if "year" in b.coords and "year" in capacity.coords:
        b = b.drop_vars("year")

    # If there are timeslices, we need to find the one where more capacity is needed to
    # meet the demand which would be a combination of a high demand and a low
    # utilization factor.
    if "timeslice" in b.dims or "timeslice" in capacity.dims:
        ratio = b / capacity
        ts_index = ratio.min("replacement").argmax("timeslice")
        b = b.isel(timeslice=ts_index)
        capacity = capacity.isel(timeslice=ts_index)

    # An adjustment is required to account for technologies that have multiple output
    # commodities
    b = modify_dlc(technologies=capacity, demand=b)

    return xr.Dataset(
        dict(capacity=capacity, b=b),
        attrs=dict(kind=ConstraintKind.UPPER_BOUND),
    )


def modify_dlc(technologies: xr.DataArray, demand: xr.DataArray) -> xr.DataArray:
    """Modifies DLC constraint to account for techs with multiple output commodities.

    Adjusts the commodity-level DLC based on the commodity output ratios of the
    available technologies, to allow for appropriate production of side-products.

    Args:
        technologies: DataArray with dimensions "commodity" and "replacement". This
            defines the fixed commodity outputs for each potential replacement
            technology.
        demand: DataArray with dimension "commodity", which defines the demand for each
            commodity.

    Returns:
        DataArray with dimension "commodity", which defines the new demand-limiting
        capacity constraint for each commodity.

    Example:
        Let's consider a simple example of a refinery sector with two alternative
        technologies that each produce two commodities: gasoline and diesel.

        We define the technologies DataArray as follows:
        >>> import xarray as xr
        >>> technologies = xr.DataArray(
        ...     data=[[1, 5], [0.5, 1]],
        ...     dims=['replacement', 'commodity'],
        ...     coords={'replacement': ['technology1', 'technology2'],
        ...             'commodity': ['gasoline', 'diesel']},
        ... )

        technology1 produces 1 unit of gasoline and 5 units of diesel (per unit of
        activity), whereas technology2 produces 0.5 units of gasoline and 1 unit of
        diesel.

        In this scenario, let's also define the demand for gasoline and diesel as
        follows (1 unit of demand for gasoline and 0 units for diesel):
        >>> demand = xr.DataArray(
        ...     data=[1, 0],
        ...     dims=['commodity'],
        ...     coords={'commodity': ['gasoline', 'diesel']},
        ... )

        The aim of the demand-limiting capacity (DLC) constraint is to limit the
        capacity of each technology so that supply is sufficient to meet the demand for
        each commodity, and no more.

        However, in this case we have a problem. The demand for gasoline can be met by
        either technology1 or technology2 (as both produce gasoline), but doing so would
        require producing up to 5 units of diesel (if all demand was met by
        technology1), which would exceed the diesel demand (0). Therefore, to allow the
        model to meet the demand for gasoline via either technology, we must relax the
        DLC constraint on diesel (to 5 units).

        In general, for an arbitrary set of technologies and commodity demands, the
        DLC of each commodity needs to be sufficiently high to permit any technology to
        act in service of any appropriate commodity demand, and no higher.

        The first step is to calculate the commodity output ratios for each technology:
        >>> output_ratios = technologies.rename({"commodity": "commodity2"}) / technologies
        >>> output_ratios
        <xarray.DataArray (replacement: 2, commodity2: 2, commodity: 2)> Size: 64B
        array([[[1. , 0.2],
                [5. , 1. ]],
        <BLANKLINE>
               [[1. , 0.5],
                [2. , 1. ]]])
        Coordinates:
          * replacement  (replacement) <U11 88B 'technology1' 'technology2'
          * commodity2   (commodity2) <U8 64B 'gasoline' 'diesel'
          * commodity    (commodity) <U8 64B 'gasoline' 'diesel'

        We introduce the dimension "commodity2" to compare the outputs of each commodity
        against every other commodity. For example, for technology1, producing 1 unit of
        gasoline leads to 5 units of diesel, whereas producing 1 unit of diesel leads to
        0.2 units of gasoline.

        Multiplying these output ratios by the demand, we get the full outputs that each
        technology would produce whilst acting in service of each commodity-level
        demand:
        >>> outputs = output_ratios * demand
        >>> outputs
        <xarray.DataArray (replacement: 2, commodity2: 2, commodity: 2)> Size: 64B
        array([[[1., 0.],
                [5., 0.]],
        <BLANKLINE>
               [[1., 0.],
                [2., 0.]]])
        Coordinates:
          * replacement  (replacement) <U11 88B 'technology1' 'technology2'
          * commodity2   (commodity2) <U8 64B 'gasoline' 'diesel'
          * commodity    (commodity) <U8 64B 'gasoline' 'diesel'

        In this case, meeting the gasoline demand with technology1 would require
        producing 1 unit of gasoline and 5 units of diesel, whereas meeting the gasoline
        demand with technology2 would require producing 1 unit of gasoline and 2 units
        of diesel. Since there is no diesel demand, all values for commodity = "diesel"
        are zero.

        Then, taking a maximum over the "commodity" dimension, we get the maximum
        potential outputs of each technology:
        >>> max_outputs = outputs.max("commodity")
        >>> max_outputs
        <xarray.DataArray (replacement: 2, commodity2: 2)> Size: 32B
        array([[1., 5.],
               [1., 2.]])
        Coordinates:
          * replacement  (replacement) <U11 88B 'technology1' 'technology2'
          * commodity2   (commodity2) <U8 64B 'gasoline' 'diesel'

        In this case, this is just the outputs of each technology when acting in service
        of the gasoline demand.

        Finally, summing over the "replacement" dimension, we get the maximum potential
        outputs of each commodity:
        >>> dlc = max_outputs.max("replacement").rename({"commodity2": "commodity"})
        >>> dlc
        <xarray.DataArray (commodity: 2)> Size: 16B
        array([1., 5.])
        Coordinates:
          * commodity  (commodity) <U8 64B 'gasoline' 'diesel'

        In this case, we get the maximum potential production of diesel as 5 units,
        which would occur as a side-product when technology1 is acting in service of the
        gasoline demand. This becomes the new DLC constraint.

        Putting this all together:
        >>> from muse.constraints import modify_dlc
        >>> modify_dlc(technologies, demand)
        <xarray.DataArray (commodity: 2)> Size: 16B
        array([1., 5.])
        Coordinates:
          * commodity  (commodity) <U8 64B 'gasoline' 'diesel'
    """  # noqa: E501
    # Calculate commodity output ratios for each technology
    output_ratios = technologies.rename({"commodity": "commodity2"}) / technologies
    output_ratios = output_ratios.where(np.isfinite(output_ratios), 0)  # this is
    # necessary for technologies that do not produce every commodity, which would lead
    # to an "infinite" ratio between commodities

    # Calculate the full outputs of each technology acting in service of each commodity
    # demand
    outputs = output_ratios * demand

    # Maximum potential outputs for each technology
    max_outputs = outputs.max("commodity")

    # Maximum potential production of each commodity -> demand-limiting capacity
    b = max_outputs.max("replacement").rename({"commodity2": "commodity"})
    return b


@register_constraints
def minimum_service(
    demand: xr.DataArray,
    capacity: xr.DataArray,
    search_space: xr.DataArray,
    technologies: xr.Dataset,
    *,
    timeslice_level: str | None = None,
    **kwargs,
) -> Constraint | None:
    """Constructs constraint between capacity and minimum service."""
    from xarray import ones_like, zeros_like

    if np.all(technologies["minimum_service_factor"] == 0):
        return None

    kwargs = dict(commodity=demand.commodity)
    if "region" in search_space.coords and "region" in technologies.dims:
        kwargs["region"] = search_space.region
    techs = (
        technologies[["fixed_outputs", "minimum_service_factor"]]
        .sel(**kwargs)
        .rename(technology="replacement")
    )
    capacity = distribute_timeslice(
        techs.fixed_outputs, level=timeslice_level
    ) * broadcast_timeslice(techs.minimum_service_factor, level=timeslice_level)
    if "asset" not in capacity.dims and "asset" in search_space.dims:
        capacity = capacity.expand_dims(asset=search_space.asset)
    production = ones_like(capacity)
    b = zeros_like(production)

    return xr.Dataset(
        dict(capacity=-capacity, production=production, b=b),
        attrs=dict(kind=ConstraintKind.LOWER_BOUND),
    )



import pytest

from muse import examples
from muse.readers.csv import (
    read_agent_parameters_csv,
    read_existing_trade_csv,
    read_global_commodities_csv,
    read_initial_capacity_csv,
    read_macro_drivers_csv,
    read_presets_csv,
    read_projections_csv,
    read_regression_parameters_csv,
    read_technodata_timeslices_csv,
    read_technodictionary_csv,
    read_timeslice_shares_csv,
    read_trade_technodata_csv,
)


@pytest.fixture
def model_path(tmp_path):
    """Creates temporary folder containing the default model."""
    examples.copy_model(name="default", path=tmp_path)
    return tmp_path / "model"


@pytest.fixture
def timeslice_model_path(tmp_path):
    """Creates temporary folder containing the default model."""
    examples.copy_model(name="default_timeslice", path=tmp_path)
    return tmp_path / "model"


@pytest.fixture
def trade_model_path(tmp_path):
    """Creates temporary folder containing the default model."""
    examples.copy_model(name="trade", path=tmp_path)
    return tmp_path / "model"


@pytest.fixture
def correlation_model_path(tmp_path):
    """Creates temporary folder containing the correlation model."""
    examples.copy_model(name="default_correlation", path=tmp_path)
    return tmp_path / "model"


def test_read_technodictionary_csv(model_path):
    """Test reading the technodictionary CSV file."""
    technodictionary_path = model_path / "power" / "Technodata.csv"
    technodictionary_df = read_technodictionary_csv(technodictionary_path)
    assert technodictionary_df is not None
    mandatory_columns = {
        "cap_exp",
        "region",
        "var_par",
        "fix_exp",
        "interest_rate",
        "utilization_factor",
        "minimum_service_factor",
        "year",
        "cap_par",
        "var_exp",
        "technology",
        "technical_life",
        "fix_par",
    }
    extra_columns = {
        "agent1",
        "max_capacity_growth",
        "total_capacity_limit",
        "max_capacity_addition",
    }
    assert set(technodictionary_df.columns) == mandatory_columns | extra_columns


def test_read_technodata_timeslices_csv(timeslice_model_path):
    """Test reading the technodata timeslices CSV file."""
    timeslices_path = timeslice_model_path / "power" / "TechnodataTimeslices.csv"
    timeslices_df = read_technodata_timeslices_csv(timeslices_path)
    mandatory_columns = {
        "utilization_factor",
        "technology",
        "minimum_service_factor",
        "region",
        "year",
    }
    extra_columns = {
        "month",
        "hour",
        "day",
    }
    assert set(timeslices_df.columns) == mandatory_columns | extra_columns


def test_read_initial_capacity_csv(model_path):
    """Test reading the initial capacity CSV file."""
    capacity_path = model_path / "power" / "ExistingCapacity.csv"
    capacity_df = read_initial_capacity_csv(capacity_path)
    mandatory_columns = {
        "region",
        "technology",
    }
    extra_columns = {
        "2025",
        "2030",
        "2035",
        "2050",
        "2045",
        "2040",
        "2020",
    }
    assert set(capacity_df.columns) == mandatory_columns | extra_columns


def test_read_global_commodities_csv(model_path):
    """Test reading the global commodities CSV file."""
    commodities_path = model_path / "GlobalCommodities.csv"
    commodities_df = read_global_commodities_csv(commodities_path)
    mandatory_columns = {
        "commodity",
        "commodity_type",
    }
    extra_columns = {"unit"}
    assert set(commodities_df.columns) == mandatory_columns | extra_columns


def test_read_timeslice_shares_csv(correlation_model_path):
    """Test reading the timeslice shares CSV file."""
    shares_path = (
        correlation_model_path / "residential_presets" / "TimesliceSharepreset.csv"
    )
    shares_df = read_timeslice_shares_csv(shares_path)
    mandatory_columns = {
        "region",
        "timeslice",
    }
    extra_columns = {
        "electricity",
        "heat",
        "wind",
        "gas",
        "CO2f",
    }
    assert set(shares_df.columns) == mandatory_columns | extra_columns


def test_read_agent_parameters_csv(model_path):
    """Test reading the agent parameters CSV file."""
    agents_path = model_path / "Agents.csv"
    agents_df = read_agent_parameters_csv(agents_path)
    mandatory_columns = {
        "search_rule",
        "quantity",
        "region",
        "type",
        "name",
        "agent_share",
        "decision_method",
    }
    extra_columns = {
        "obj_sort1",
        "objective1",
        "obj_data1",
    }
    assert set(agents_df.columns) == mandatory_columns | extra_columns


def test_read_macro_drivers_csv(correlation_model_path):
    """Test reading the macro drivers CSV file."""
    macro_path = correlation_model_path / "residential_presets" / "Macrodrivers.csv"
    macro_df = read_macro_drivers_csv(macro_path)
    mandatory_columns = {
        "region",
        "variable",
    }
    extra_columns = {
        *[str(year) for year in range(2010, 2111)],
    }
    assert set(macro_df.columns) == mandatory_columns | extra_columns

    assert "Population" in macro_df["variable"].values
    assert "GDP|PPP" in macro_df["variable"].values


def test_read_projections_csv(model_path):
    """Test reading the projections CSV file."""
    projections_path = model_path / "Projections.csv"
    projections_df = read_projections_csv(projections_path)
    mandatory_columns = {
        "year",
        "attribute",
        "region",
    }
    extra_columns = {
        "CO2f",
        "electricity",
        "gas",
    }
    assert set(projections_df.columns) == mandatory_columns | extra_columns


def test_read_regression_parameters_csv(correlation_model_path):
    """Test reading the regression parameters CSV file."""
    regression_path = (
        correlation_model_path / "residential_presets" / "regressionparameters.csv"
    )
    regression_df = read_regression_parameters_csv(regression_path)
    mandatory_columns = {
        "sector",
        "region",
        "function_type",
        "coeff",
    }
    extra_columns = {
        "CO2f",
        "electricity",
        "gas",
        "heat",
    }
    assert set(regression_df.columns) == mandatory_columns | extra_columns


def test_read_presets_csv(model_path):
    """Test reading the presets CSV files."""
    presets_path = model_path / "residential_presets" / "Residential2020Consumption.csv"
    presets_df = read_presets_csv(presets_path)

    mandatory_columns = {
        "region",
        "timeslice",
    }
    extra_columns = {
        "heat",
    }
    assert set(presets_df.columns) == mandatory_columns | extra_columns


def test_read_existing_trade_csv(trade_model_path):
    """Test reading the existing trade CSV file."""
    trade_path = trade_model_path / "power" / "ExistingTrade.csv"
    trade_df = read_existing_trade_csv(trade_path)
    mandatory_columns = {
        "region",
        "technology",
        "year",
    }
    extra_columns = {"r1", "r2"}
    assert set(trade_df.columns) == mandatory_columns | extra_columns


def test_read_trade_technodata(trade_model_path):
    """Test reading the trade technodata CSV file."""
    trade_path = trade_model_path / "power" / "TradeTechnodata.csv"
    trade_df = read_trade_technodata_csv(trade_path)
    mandatory_columns = {"technology", "region", "parameter"}
    extra_columns = {"r1", "r2"}
    assert set(trade_df.columns) == mandatory_columns | extra_columns



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config
# -- Project information -----------------------------------------------------

from importlib.metadata import version as get_version

from docutils import nodes
from docutils.parsers.rst import roles

project = "MUSE"
copyright = "2024, Imperial College London"
author = "Imperial College London"
release = get_version("MUSE_OS")
version = ".".join(release.split(".")[:2])

# -- General configuration ---------------------------------------------------

master_doc = "index"
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.coverage",
    "sphinx.ext.napoleon",
    "nbsphinx",
    "myst_parser",
    "sphinx.ext.mathjax",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.viewcode",
    "sphinxcontrib.bibtex",
    "IPython.sphinxext.ipython_console_highlighting",
    "IPython.sphinxext.ipython_directive",
    "sphinx.ext.graphviz",
]
source_suffix = {".rst": "restructuredtext", ".txt": "markdown", ".md": "markdown"}
templates_path = ["_templates"]
exclude_patterns = [
    "build",
    "**.ipynb_checkpoints",
    "**/ResidentialBracket*.txt",
    "_sources/*",
    "_build/*",
    "tutorial-code/*",
]

modindex_common_prefix = ["muse"]
autodoc_typehints = "none"
add_module_names = False
nbsphinx_allow_errors = True
autosectionlabel_prefix_document = True
nitpicky = True

# Suppress warnings for documents not included in any toctree (e.g. release notes)
suppress_warnings = ["toc.not_included"]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("http://docs.scipy.org/doc/numpy/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "xarray": ("https://xarray.pydata.org/en/stable/", None),
}


bibtex_bibfiles: list[str] = []

# -- GraphViz configuration ----------------------------------
graphviz_output_format = "svg"

# -- Options for HTML output -------------------------------------------------

html_theme = "sphinx_rtd_theme"

# -- Render GitHub links -------------------------------------------------


def github_role(name, rawtext, text, lineno, inliner, options=None, content=None):
    """Creates links to issues/pull requests on the MUSE_OS GitHub site.

    To use in markdown:
    {github}`ISSUE_NUMBER` (e.g. {github}`123`)

    To use in rst:
    :github:`ISSUE_NUMBER` (e.g. :github:`123`)

    In both cases this will create a clickable link (visible as #123) to the relevant
    GitHub page (i.e. https://github.com/EnergySystemsModellingLab/MUSE_OS/issues/123)

    The base URL is for the issues page, but this will also work for pull requests and
    discussions, as GitHub will automatically redirect to the appropriate page.
    """
    base_url = "https://github.com/EnergySystemsModellingLab/MUSE_OS/issues/"
    url = f"{base_url}{text}"
    node = nodes.reference(rawtext, f"#{text}", refuri=url, **(options or {}))
    return [node], []


roles.register_canonical_role("github", github_role)



"""Test agent interactions."""

from collections import namedtuple
from itertools import chain

import pytest
from numpy.random import choice
from pytest import fixture, mark

from muse.errors import NoInteractionsFound
from muse.interactions import (
    agents_groupby,
    factory,
    new_to_retro_net,
    register_agent_interaction,
)


@fixture
def agents():
    Agent = namedtuple("Agent", ["region", "name", "category", "assets"])
    sample_data = {
        "regions": ["Area52", "Bermuda Triangle", "City of London"],
        "names": ["John", "Joe", "Jill"],
        "categories": ["yup", "nope"],
    }

    # Generate unique combinations of region, name, and category
    combinations = {
        tuple(items)
        for items in zip(*(choice(data, 40) for data in sample_data.values()))
    }

    return [Agent(*combo, [None]) for combo in combinations]


def test_groupby(agents):
    grouped = agents_groupby(agents, ("category", "name"))

    # Verify group sizes and agent preservation
    assert sum(len(group) for group in grouped.values()) == len(agents)
    assert set(map(id, chain(*grouped.values()))) == set(map(id, agents))

    # Verify correct grouping keys and contents
    assert set(grouped.keys()) == {(a.category, a.name) for a in agents}
    for (cat, name), group in grouped.items():
        assert all(a.category == cat and a.name == name for a in group)


def test_new_to_retro_net(agents):
    net = new_to_retro_net(agents, "nope")

    # Verify network properties
    assert sum(len(group) for group in net) <= len(agents)
    assert set(map(id, chain(*net))).issubset(map(id, agents))
    assert all(len(list(group)) == 2 for group in net)

    # Verify group constraints
    for group in net:
        assert len({a.region for a in group}) == 1
        assert len({a.name for a in group}) == 1
        categories = [a.category for a in group]
        yup_index = categories.index("yup") if "yup" in categories else 0
        assert "nope" not in categories[yup_index:]
        assert "yup" not in categories[:yup_index]


@mark.usefixtures("save_registries")
def test_compute_interactions(agents):
    @register_agent_interaction
    def dummy_interaction(a, b):
        assert all(agent.assets[0] is None for agent in (a, b))
        a.assets[0], b.assets[0] = b, a

    interactions = factory([("new_to_retro", "dummy_interaction")])
    interactions(agents)

    # Check unmatched agents
    unmatched = [agent for agent in agents if agent.assets[0] is None]
    assert len({(a.region, a.name) for a in unmatched}) == len(unmatched)

    # Check matched agents
    matched = [agent for agent in agents if agent.assets[0] is not None]
    assert len(new_to_retro_net(agents)) == 0 or matched
    assert all(agent.assets[0].assets[0] is agent for agent in matched)

    # Test that NoInteractionsFound is raised when all agents are 'nope' category
    nope_agents = [a for a in agents if a.category == "nope"]
    with pytest.raises(NoInteractionsFound):
        interactions(nope_agents)



r"""Collection of demand-matching algorithms.

At it's simplest, the demand matching algorithm solves the following problem,

- given a demand for a commodity :math:`D_d`, with :math:`d\in\mathcal{D}`
- given processes to supply these commodities, with an associated cost per process,
  :math:`C_{d, i}`, with :math:`i\in\mathcal{I}`

Match demand and supply while minimizing the associated cost.

.. math::

   \min_{X} \sum_{d, i} C_{d,i} X_{d, i}

   X_{d, i} \geq 0

   \sum_o X_o \geq D_d

The basic algorithm proceeds as follows:

#. sort all costs :math:`C_{d, i}` across both :math:`d` and :math:`i`

#. for each cost :math:`c_0` in order:

    #. find the set of indices :math:`\mathcal{C}\subseteq\mathcal{D}\cup\mathcal{I}`
       for which

        .. math::

            \forall (d, i) \in \mathcal{C}\quad C_{d, i} == c_0

    #. determine the partial result for the current cost

        .. math::

            \forall (d, i) \in \mathcal{C}\quad X_{d, i} = \frac{D_d}{|i\in\mathcal{C}|}

        Where :math:`|i\in\mathcal{C}|` indicates the number of indices :math:`i` in
        :math:`\mathcal{C}`.

However, in practice, the problem to solve often contains constraints, e.g. a constraint
on production :math:`\sum_d X_{d, i} \leq M_i`. The algorithms in this module try and
solve these constrained problems one way or another.
"""

from __future__ import annotations

__all__ = ["demand_matching"]


import pandas as pd
from xarray import DataArray

from muse.timeslices import drop_timeslice


def demand_matching(
    demand: DataArray,
    cost: DataArray,
    *constraints: DataArray,
    protected_dims: set | None = None,
) -> DataArray:
    r"""Demand matching over heterogeneous dimensions.

    This algorithm enables demand matching while enforcing constraints on how much an
    asset can produce. Any set of dimensions can be matched. The algorithm is general
    with respect to the dimensions in demand and cost. It also enforces constraints over
    sets of indices.


    .. math::

        \min_{X} \sum_{d, i} C_{d, i} X_{d, i}

        X_{d, i} \geq 0

        \sum_i X_{d, i} \geq D_d

        M_{(d, i) \in \mathcal{R}^{(\alpha)}}^{(\alpha)}
            \geq \sum_{(d, i)\notin\mathcal{R}^{(\alpha)}} X_{d, i}

    Where :math:`\alpha` is an index running over constraints,
    :math:`\mathcal{R}^{(\alpha)}\subseteq\mathcal{D}\cup\mathcal{I}` is a subset of
    indices.

    The algorithm proceeds as described in :py:mod:`muse.demand_matching`.
    However, an extra step is added to ensure that the solutions falls within the
    convex-hull formed by the constraints. This projects the current solution onto the
    constraint. Hence, the solution will depend on the order in which the constraints
    are given.

    #. sort all costs :math:`C_{d, m}` across both :math:`d` and :math:`m`

    #. for each cost :math:`c_0` in order:

        #. find the set of indices :math:`\mathcal{C}`

            .. math::

                \mathcal{C}\subseteq\mathcal{D}\cup\mathcal{I}

                \forall (d, i) \in \mathcal{C}\quad C_{d, i} == c_0

        #. determine an interim partial result for the current cost

            .. math::

                \forall (d, i) \in \mathcal{C}\quad
                \delta X_{d, i} = \frac{1}{|i\in\mathcal{C}|}\left(
                    D_d - \sum_{j \in \mathcal{I}} X_{d, j}\right)

            Where :math:`|i\in\mathcal{C}|` indicates the number of :math:`i` indices in
            :math:`\mathcal{C}`. The expression in the parenthesis is the currently
            unserviced demand.

        #. Loop over each constraint :math:`\alpha`. Below we drop the index
           :math:`\alpha` over constraints for simplicity.

            #. Determine the excess over the constraint:

                .. math::

                    E_{(d, i) \in \mathcal{R}} = \max\left\{
                        0,
                        \sum_{(d, i)\notin\mathcal{R}}\left(
                            X_{d, i} + \delta X_{d, i}
                        \right) - M_{(d, i) \in \mathcal{R}}
                    \right\}

            #. Correct :math:`\delta X` as follows:

                .. math::

                    \forall (d, i) \in \mathcal{C}\cap\mathcal{R}\quad
                    \delta X\prime_{d, i} =
                        E_{(d, i)}
                        \frac{\delta X_{(d, i)}}{
                            \sum_{(e, j)\in \mathcal{C}\cap\mathcal{R}} \delta X_{(e,j)}
                        }


                    \forall (d, i) \notin \mathcal{R}, (d, i)\in\mathcal{C} \quad
                    \delta X\prime_{d, i} = 0

            #. Set :math:`\delta X = \max(0, \delta X - \delta X\prime)`


    A more complex problem would see independent dimensions for each quantity. In that,
    case we can reduce to the original problem as shown here

    .. math::

        C_{d, i, c} = \min_cC\prime_{d, i, c}

        D_d = \sum_{d\prime} D\prime_{d, d\prime}

        M_r = \sum_m M\prime_{r, m}

        X_{d, d\prime, i, m, c} =
            \left(C\prime_{d, i, c} == C_{d, i}\right)
            \frac{M\prime_{r, m}}{M_r} \frac{D\prime_{d, d\prime}}{D_d} X_{d, i}

    A dimension could be shared by all quantities, in which case each point along that
    dimension is treated as independent.

    Similarly, if a dimension is shared only by the demand and a constraint but not by
    the cost, then the problem can be reduced a set of problems independent along that
    direction.

    Arguments:
        demand: Demand to match with production. It should have the same physical units
            as `max_production`.
        cost: Cost to minimize while fulfilling the demand.
        *constraints: each item is a separate constraint :math:`M_r`.
        protected_dims: Dimensions that will not be modified

    Returns:
        An array with the joint dimensionality of `max_production`, `cost`, and
        `demand`, containing the supply that fulfills the demand. The units of this
        supply are the same as `demand` and `max_production`.
    """
    from pandas import MultiIndex
    from xarray import Dataset

    if protected_dims is None:
        protected_dims = set()

    # demand has extra dimensions unique to it
    extra_dims = set(demand.dims).difference(
        cost.dims, *(cons.dims for cons in constraints)
    )

    if extra_dims:
        summed_demand = demand.sum(extra_dims)
        demand_share = (demand / summed_demand).where(demand > 0, 0)
        result = demand_matching(summed_demand, cost, *constraints)
        return demand_share * result

    # a constraint has dimensions unique to it
    for i, constraint in enumerate(constraints):
        others = [cons for j, cons in enumerate(constraints) if j != i]
        extra_dims = set(constraint.dims).difference(
            demand.dims, cost.dims, *(cons.dims for cons in others), protected_dims
        )
        if extra_dims:
            summed_constraint = constraint.sum(extra_dims)
            share = (constraint / summed_constraint).where(constraint > 0, 0)
            others.insert(i, summed_constraint)
            result = demand_matching(demand, cost, *others)
            return share * result

    # cost has extra dimensions unique to it
    extra_dims = set(cost.dims).difference(
        demand.dims, *(cons.dims for cons in constraints), protected_dims
    )
    if extra_dims:
        mincost = cost.min(extra_dims)
        result = demand_matching(demand, mincost, *constraints)
        structure = cost == mincost
        cost_share = (1 / structure.sum(extra_dims)).where(structure, 0)
        return result * cost_share

    # missing coordinates in max_production
    # add a fake coordinate so that the individual items in coordinate-less dimension
    # can be addressed at the end of the main loop.
    ds = Dataset(
        {
            **{f"constraint{i}": cons for i, cons in enumerate(constraints)},
            "cost": cost,
            "demand": demand,
        }
    )

    nocoords = set(ds.dims).difference(ds.coords.keys())

    if nocoords:
        for coord in nocoords:
            ds.coords[coord] = coord, ds.coords[coord].values
        result = demand_matching(  # type: ignore
            ds.demand, ds.cost, *(ds[f"constraint{i}"] for i in range(len(constraints)))
        )
        return result.drop_vars(nocoords)

    # multi-index dimensions seem to make life complicated for groupby
    # so drop them for the duration of the call.
    # see https://github.com/pydata/xarray/issues/1603
    multics = {
        k: ds.coords[k] for k in ds.dims if isinstance(ds.get_index(k), MultiIndex)
    }

    if len(multics) > 0:
        for k in multics:
            ds = drop_timeslice(ds)
            ds[k] = pd.Index(constraint.get_index(k), tupleize_cols=False)
        result = demand_matching(  # type: ignore
            ds.demand, ds.cost, *(ds[f"constraint{i}"] for i in range(len(constraints)))
        )
        for k, v in multics.items():
            result.coords[k] = v
        return result
    return _demand_matching_impl(demand, cost, *constraints)


def _demand_matching_impl(
    demand: DataArray, cost: DataArray, *constraints: DataArray
) -> DataArray:
    """Implementation of demand matching.

    It is expected that demand does not have dimension unique to itself, and that all
    dimensions have coordinates. Input sanitization is performed in `demand_matching`
    proper.
    """
    from numpy import isnan, prod
    from xarray import Dataset, align, full_like

    assert not set(demand.dims).difference(
        cost.dims, *(cons.dims for cons in constraints)
    )

    result = full_like(sum(constraints) + demand + cost, 0)  # type: ignore
    names = [f"constraint{i}" for i in range(len(constraints))]
    data = Dataset(
        {
            "cost": cost,
            "demand": demand.copy(deep=True),
            **dict(zip(names, constraints)),
        }
    )

    def expand_dims(x, like):
        """Add extra dims that are in ``like``."""
        N = max(1, prod([len(like[d]) for d in like.dims if d not in x.dims]))
        b = (x / N).expand_dims(**{d: like[d] for d in like.dims if d not in x.dims})
        return b

    def remove_dims(x, to):
        """Remove extra dims that are not in ``to``."""
        extras = set(x.dims).difference(to.dims)
        a = x.sum(extras) / (~isnan(x)).sum(extras)
        return a

    idims = [dim for dim in result.dims if dim not in demand.dims]
    for _, same_cost in data.groupby("cost") if cost.dims else [(cost, data)]:
        current = same_cost.drop_vars("cost").unstack()
        assert (~isnan(result)).all()
        delta_x = expand_dims(
            (remove_dims(current.demand, demand) - result.sum(idims)).clip(0), current
        )
        for cname in names:
            constraint = remove_dims(current[cname], data[cname])
            delta_x, constraint = align(delta_x, constraint, join="outer", fill_value=0)
            excess = (
                result.sum(set(data.dims).difference(data[cname].dims))
                + delta_x.sum(set(data.dims).difference(data[cname].dims))
                - constraint
            ).clip(min=0)
            excess_share = (
                excess
                * (
                    delta_x / delta_x.sum(set(delta_x.dims).difference(constraint.dims))
                ).fillna(0)
            ).fillna(0)
            delta_x = (expand_dims(delta_x, excess_share) - excess_share).clip(0)

        result = sum(align(result, delta_x.fillna(0), fill_value=0, join="left"))

    return result



from pathlib import Path

from conftest import chdir
from pytest import mark, xfail

from muse.examples import AVAILABLE_EXAMPLES


@mark.regression
@mark.example
@mark.parametrize("model", AVAILABLE_EXAMPLES)
def test_fullsim_regression(model, tmp_path, compare_dirs):
    from warnings import simplefilter

    from pandas.errors import DtypeWarning

    from muse.examples import copy_model
    from muse.mca import MCA

    # fail the test if this warning crops up
    simplefilter("error", DtypeWarning)

    # Copy the data to tmp_path
    model_path = copy_model(name=model, path=tmp_path)

    # main() will output to cwd
    with chdir(tmp_path):
        MCA.factory(model_path / "settings.toml").run()

    compare_dirs(
        tmp_path / "Results",
        Path(__file__).parent / "example_outputs" / model.replace("-", "_") / "Results",
        rtol=1e-4,
        atol=1e-7,
    )


def available_tutorials():
    base_path = Path(__file__).parent.parent / "docs" / "tutorial-code"
    return [p.parent for p in base_path.rglob("settings.toml")]


@mark.regression
@mark.tutorial
@mark.parametrize("tutorial_path", available_tutorials())
def test_tutorial_regression(tutorial_path, tmp_path, compare_dirs):
    import shutil
    from warnings import simplefilter

    from pandas.errors import DtypeWarning

    from muse.mca import MCA

    # Mark as xfail for a specific tutorial
    if "modify-time-framework" in str(tutorial_path):
        xfail(reason="Known issue with this tutorial (#371)")

    # fail the test if this warning crops up
    simplefilter("error", DtypeWarning)

    # Copy the data to tmp_path
    shutil.copytree(tutorial_path, tmp_path, dirs_exist_ok=True)

    # Get the toml file to run. There should be only one settings file
    settings = list(tmp_path.glob("*.toml"))
    assert len(settings) == 1

    # Rename existing results directory for comparison
    expected = tmp_path / "Expected"
    (tmp_path / "Results").rename(expected)

    # main() will output to cwd
    # Change working directory to tmp_path
    with chdir(tmp_path):
        MCA.factory(settings[0]).run()

    compare_dirs(
        tmp_path / "Results",
        expected,
        rtol=1e-4,
        atol=1e-7,
    )



import os
import shutil
from pathlib import Path

import pandas as pd

from muse import examples
from muse.wizard import (
    add_new_commodity,
    add_new_process,
    add_technodata_for_new_year,
    modify_toml,
)

parent_path = Path(__file__).parent


def generate_model_1() -> None:
    """Generates the first model for tutorial 1.

    Adds solarPV to the default model.

    """
    model_name = "1-introduction"
    model_path = parent_path / model_name
    if model_path.exists():
        shutil.rmtree(model_path)

    # Starting point: copy default model
    examples.copy_model(name="default", path=parent_path, overwrite=True)
    os.rename(parent_path / "model", model_path)

    # Copy wind commodity in power sector -> solar
    add_new_commodity(model_path, "solar", "power", "wind")

    # Copy windturbine process in power sector -> solarPV
    add_new_process(model_path, "solarPV", "power", "windturbine")

    # Modify input commodities
    commin_file = model_path / "power/CommIn.csv"
    df = pd.read_csv(commin_file)
    df.loc[(df["technology"] == "solarPV"), "solar"] = 1
    df.loc[(df["technology"] == "solarPV"), "wind"] = 0
    df.loc[(df["technology"] == "windturbine"), "solar"] = 0
    df.fillna(0, inplace=True)
    df.to_csv(commin_file, index=False)

    # Modify technodata for solarPV
    technodata_file = model_path / "power/Technodata.csv"
    df = pd.read_csv(technodata_file)
    df.loc[df["technology"] == "solarPV", "cap_par"] = 30
    df.to_csv(technodata_file, index=False)

    # Add solar to excluded commodities
    settings_file = model_path / "settings.toml"
    modify_toml(
        settings_file, lambda settings: settings["excluded_commodities"].append("solar")
    )


def generate_model_2() -> None:
    """Generates the second model for tutorial 1.

    Changes the price of solar between 2020 and 2040.

    """
    model_name = "2-scenario"
    model_path = parent_path / model_name
    if model_path.exists():
        shutil.rmtree(model_path)

    # Starting point: copy model from 1-introduction
    shutil.copytree(parent_path / "1-introduction", model_path)
    if (model_path / "Results").exists():
        shutil.rmtree(model_path / "Results")

    # Copy technodata for 2020 -> 2040
    add_technodata_for_new_year(model_path, 2040, "power", 2020)

    # Modify cap_par for solarPV
    technodata_file = model_path / "power/Technodata.csv"
    df = pd.read_csv(technodata_file)
    df.loc[(df["technology"] == "solarPV") & (df["year"] == 2020), "cap_par"] = 40
    df.loc[(df["technology"] == "solarPV") & (df["year"] == 2040), "cap_par"] = 30
    df.to_csv(technodata_file, index=False)


if __name__ == "__main__":
    generate_model_1()
    generate_model_2()



from pytest import fixture


@fixture(autouse=True)
def add_np(doctest_namespace):
    import numpy
    import pandas
    import xarray

    doctest_namespace["np"] = numpy
    doctest_namespace["xr"] = xarray
    doctest_namespace["pd"] = pandas



from unittest.mock import patch

import numpy as np
from pytest import approx, fixture


@fixture
def linear_sampling():
    from muse.carbon_budget import linear_fun

    a = 0.01
    budget = 420000
    carbon_price = 0.02
    b = (carbon_price - a) / budget

    emissions = np.random.uniform(low=budget * 0.75, high=budget * 1.25, size=4)
    prices = linear_fun(emissions, a, b)

    return prices, emissions, budget, a, b


@fixture
def exp_sampling():
    from muse.carbon_budget import exponential_fun

    c = 0.01
    budget = 420000
    b = 1 / budget
    carbon_price = 0.02
    a = (carbon_price - c) / np.exp(-1)

    emissions = np.random.uniform(low=budget * 0.75, high=budget * 1.25, size=4)
    prices = exponential_fun(emissions, a, b, c)

    return prices, emissions, budget, a, b, c


def test_create_sample():
    from muse.carbon_budget import create_sample

    carbon_price = 0.02
    budget = 420000
    size = 4

    current_emissions = 300000
    expected = np.array([0.02, 0.01933333, 0.01866667, 0.018])
    sample = create_sample(carbon_price, current_emissions, budget, size)
    assert sample == approx(expected)

    current_emissions = 500000
    expected = np.array([0.02, 0.02115623, 0.02231246, 0.02346869])
    sample = create_sample(carbon_price, current_emissions, budget, size)
    assert sample == approx(expected)


def test_linear_guess_and_weights(linear_sampling):
    from muse.carbon_budget import linear_guess_and_weights

    (prices, emissions, budget, a, b) = linear_sampling

    actual_weights = np.abs(emissions - budget)

    (est_a, est_b), est_weights = linear_guess_and_weights(prices, emissions, budget)

    assert actual_weights == approx(est_weights)
    assert a == approx(est_a)
    assert b == approx(est_b)


def test_linear(linear_sampling):
    from muse.carbon_budget import linear, linear_fun

    (prices, emissions, budget, a, b) = linear_sampling

    actual_price = linear_fun(budget, a, b)
    est_price = linear(prices, emissions, budget)
    assert actual_price == approx(est_price)


def test_exp_guess_and_weights(exp_sampling):
    from muse.carbon_budget import exp_guess_and_weights

    (prices, emissions, budget, a, b, c) = exp_sampling

    actual_weights = np.abs(emissions - budget)

    (est_a, est_b, est_c), est_weights = exp_guess_and_weights(
        prices, emissions, budget
    )

    assert actual_weights == approx(est_weights)
    assert a == approx(est_a, rel=1e3)
    assert b == approx(est_b, rel=1e3)
    assert c == approx(est_c, rel=1e3)


def test_exponential(exp_sampling):
    from muse.carbon_budget import exponential, exponential_fun

    (prices, emissions, budget, a, b, c) = exp_sampling

    actual_price = exponential_fun(budget, a, b, c)

    est_price = exponential(prices, emissions, budget)
    assert actual_price == approx(est_price)


def test_overshoot():
    from muse.carbon_budget import update_carbon_budget

    year = 0
    carbonbudget = np.array([4e6, 5e6, 6e6], dtype=int)

    emissions = int(5.3e6)
    expected = int(5.7e6)
    actual = update_carbon_budget(carbonbudget, emissions, year, over=True, under=False)
    assert expected == actual

    emissions = int(4e6)
    expected = int(6e6)
    actual = update_carbon_budget(carbonbudget, emissions, year, over=True, under=False)
    assert expected == actual


def test_undershoot():
    from muse.carbon_budget import update_carbon_budget

    year = 0
    carbonbudget = np.array([4e6, 5e6, 6e6], dtype=int)

    emissions = int(4.7e6)
    expected = int(6.3e6)
    actual = update_carbon_budget(carbonbudget, emissions, year, over=False, under=True)
    assert expected == actual

    emissions = int(6e6)
    expected = int(6e6)
    actual = update_carbon_budget(carbonbudget, emissions, year, over=False, under=True)
    assert expected == actual


@patch("muse.carbon_budget.decrease_bounds")
@patch("muse.carbon_budget.increase_bounds")
@patch("muse.carbon_budget.bisect_bounds")
@patch("muse.carbon_budget.bisect_bounds_inverted")
def test_adjust_bounds(
    mock_decrease_bounds,
    mock_increase_bounds,
    mock_bisect_bounds,
    mock_bisect_bounds_inverted,
):
    from muse.carbon_budget import adjust_bounds

    lb_price = 1.0
    ub_price = 2.0
    target = 20.0

    # Test 1: lb_price_emissions < target and ub_price_emissions < target
    emissions = {lb_price: 10.0, ub_price: 15.0}
    adjust_bounds(lb_price, ub_price, emissions, target)
    mock_decrease_bounds.asset_called_once()

    # Test 2: lb_price_emissions > target and ub_price_emissions > target
    emissions = {lb_price: 25.0, ub_price: 30.0}
    adjust_bounds(lb_price, ub_price, emissions, target)
    mock_increase_bounds.asset_called_once()

    # Test 3: lb_price_emissions > target and ub_price_emissions < target
    emissions = {lb_price: 25.0, ub_price: 15.0}
    adjust_bounds(lb_price, ub_price, emissions, target)
    mock_bisect_bounds.asset_called_once()

    # Test 4: lb_price_emissions < target and ub_price_emissions > target
    emissions = {lb_price: 10.0, ub_price: 30.0}
    adjust_bounds(lb_price, ub_price, emissions, target)
    mock_bisect_bounds_inverted.asset_called_once()


def test_decrease_bounds():
    from muse.carbon_budget import decrease_bounds

    lb_price = 1.0
    ub_price = 2.0
    emissions = {lb_price: 10.0, ub_price: 30.0}
    target = 20.0
    new_lb_price, new_ub_price = decrease_bounds(lb_price, ub_price, emissions, target)
    assert new_lb_price < lb_price
    assert new_ub_price == lb_price


def test_increase_bounds():
    from muse.carbon_budget import increase_bounds

    lb_price = 1.0
    ub_price = 2.0
    emissions = {lb_price: 10.0, ub_price: 30.0}
    target = 5.0
    new_lb_price, new_ub_price = increase_bounds(lb_price, ub_price, emissions, target)
    assert new_lb_price == ub_price
    assert new_ub_price > ub_price


def test_bisect_bounds():
    from muse.carbon_budget import bisect_bounds

    lb_price = 1.0
    ub_price = 2.0
    midpoint_price = (lb_price + ub_price) / 2
    target = 20.0

    # Test 1: midpoint emissions < target
    emissions = {lb_price: 10.0, ub_price: 30.0, midpoint_price: 12.0}
    new_lb_price, new_ub_price = bisect_bounds(lb_price, ub_price, emissions, target)
    assert new_lb_price == lb_price
    assert new_ub_price == midpoint_price

    # Test 2: midpoint emissions > target
    emissions = {lb_price: 10.0, ub_price: 30.0, midpoint_price: 22.0}
    new_lb_price, new_ub_price = bisect_bounds(lb_price, ub_price, emissions, target)
    assert new_lb_price == midpoint_price
    assert new_ub_price == ub_price


def test_bisect_bounds_inverted():
    from muse.carbon_budget import bisect_bounds_inverted

    lb_price = 1.0
    ub_price = 2.0
    midpoint_price = (lb_price + ub_price) / 2
    target = 20.0

    # Test 1: midpoint emissions < target
    emissions = {lb_price: 10.0, ub_price: 30.0, midpoint_price: 12.0}
    new_lb_price, new_ub_price = bisect_bounds_inverted(
        lb_price, ub_price, emissions, target
    )
    assert new_lb_price == midpoint_price
    assert new_ub_price == ub_price

    # Test 2: midpoint emissions > target
    emissions = {lb_price: 10.0, ub_price: 30.0, midpoint_price: 22.0}
    new_lb_price, new_ub_price = bisect_bounds_inverted(
        lb_price, ub_price, emissions, target
    )
    assert new_lb_price == lb_price
    assert new_ub_price == midpoint_price



import subprocess
import sys
from pathlib import Path

if __name__ == "__main__":
    for tutorial in sorted(Path(__file__).parent.iterdir()):
        if (tutorial / "generate_models.py").exists():
            subprocess.call([sys.executable, str(tutorial / "generate_models.py")])
